# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import abstractmethod

import torch
import torch.nn as nn
import types
import functools

from .._core.quant_dequant import QuantDequant as qdq, QuantDynamicInput
from .._core.quantized_func_wrappers import get_quantized_func_wrapper, OP_TYPE
from .quant_config import QuantMode, get_hqt_config
from ..patched_module_base import PatchedModuleBase, get_call_wrapper
from .._core.scale_handler import get_scale_dtype, ScaleFormat
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator
cur_accelerator = auto_detect_accelerator()


class BMM(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, y):
        return torch.bmm(x, y)


class Matmul(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, *args, **kwargs):
        return torch.matmul(*args, **kwargs)


class Identity(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x.clone()


class Softmax(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, dim=None, invAttnHead=None):
        return torch.softmax(x, dim)


def measure_input(input, observer):
    for i in range(len(observer)):
        observer[i].measure(input[i])


def measure_output(output, observer):
    if observer:
        for i in range(len(observer)):
            observer[i].measure(output[i])


def get_current_repr(cls_instance, *member_names):
    curr_repr = ""
    if cls_instance.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
        first_name = True
        for name in member_names:
            if not first_name:
                curr_repr += ", "
            cur_attr = getattr(cls_instance, name)
            if isinstance(cur_attr, list) and len(cur_attr) > 1:
                dtype = get_scale_dtype(cur_attr[0])
                curr_repr += f"{name} type=list of {dtype}, length={len(cur_attr)}"
            else:
                # currently, only scale is called here.
                dtype = get_scale_dtype(cur_attr)
                curr_repr += f"{name} dtype={dtype}"
            first_name = False
    return curr_repr


def extra_representation(org_repr, org_name, curr_repr):
    repr = f"original={org_name}," + (" " + org_repr + "," if org_repr != "" else "")
    return f"{repr} {curr_repr}"


def _raise_lora_layer_error(layer_class):
    raise RuntimeError(
        f"{layer_class} quantization is not supported in case of lora_layer member is not None."
        f" Can add {layer_class} to 'blocklist' field in quantization config file"
    )


class PatchedMatmul(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            self.quant_input_0 = self._mod_extra_config.inputs[0]
            self.quant_input_1 = self._mod_extra_config.inputs[1]
            if not self.use_qdq and not self.fake_quant:
                self.register_scale("scale_input", mod_extra_config.scale.inputs[0], self.scale_format)
                self.register_scale("scale_other", mod_extra_config.scale.inputs[1], self.scale_format)
                self.matmul_fp8 = get_quantized_func_wrapper(OP_TYPE.MATMUL_GEMM, self.scale_format)
                # in DPQ we want to use the scales measured in the quantization process
                if hasattr(parent, 'scale_bf16_to_fp8') and parent.scale_bf16_to_fp8 > 0:
                    self.scale_other = torch.nn.Parameter(parent.scale_bf16_to_fp8)

    def forward_quant(self, input, other):
        qinput = self.quant_input_0(input)
        qother = self.quant_input_1(other)
        output = self.matmul_fp8(qinput,
                                 qother,
                                 out_dtype=self._mod_extra_config.config_params["hp_dtype"],
                                 scale_input_inv=self.scale_input,
                                 scale_other_inv=self.scale_other)
        return output

    def forward_qdq(self, input, other):
        qinput = self.quant_input_0(input)
        qother = self.quant_input_1(other)
        output = torch.matmul(qinput, qother)
        return output

    def forward_measure(self, input, other):
        measure_input((input, other), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input, other)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_other") if not self.use_qdq else "",
        )


def init_mixture_of_experts_linears(instance):
    parent_name = instance.orig_mod_parent.__class__.__name__
    if parent_name == "MixtralBlockSparseTop2MLP" or parent_name == "GaudiDeepseekV3MLP":
        # this linear is part of MixtureOfExperts block
        # MoE linears hold the weights but their forward logic is done using the dynamic op
        # therefore no measure object is saved causing no quant object as well
        # adding a fake 0 data input to instantiate the needed measurment objects
        measure_input((torch.tensor(0),), observer=instance._mod_extra_config.inputs)


class PatchedLinearBase(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)

    # TODO [SW-224538]: Move init_linear to PatchedLinearBase __init__
    def init_linear(self, mod_extra_config):
        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            # When offloading weights to disk using device_map, the module forward is overridden.
            # __dict__.update call again overrides the PatchedLinear forward with the forward that device_map planted.
            # So need to set PatchedLinear forward to be the right forward.
            self.weight = nn.Parameter(self.weight.t().contiguous())
            self.quant_input = self._mod_extra_config.inputs[0]
            self.dequant_output = self._mod_extra_config.outputs[0]

            # When offloading weights to disk using device_map, the module forward is overridden.
            # __dict__.update call again overrides the PatchedLinear forward with the forward that device_map planted.
            # So need to set PatchedLinear forward to be the right forward.
            if self.use_qdq:
                self.dequant_weights = self._mod_extra_config.params["weight"][1]
            elif self.fake_quant:
                self.qdq_weights = self._mod_extra_config.params["weight"][0]
            else:
                self.matmul_fp8 = get_quantized_func_wrapper(OP_TYPE.LINEAR_GEMM, self.scale_format)
                # input0 is None when initializing a dynamic quantization op
                self.register_scale("scale_input", mod_extra_config.scale.inputs[0], self.scale_format)
                if isinstance(mod_extra_config.scale.params["weight"], (torch.Tensor, float)):
                    self.register_scale("scale_weight", mod_extra_config.scale.params["weight"], self.scale_format)
                elif isinstance(mod_extra_config.scale.params["weight"], dict):
                    # PCQ weight is calculated with actual weight [0] and ones [1]
                    # only ScaleFormat.CONST is supported for per-channel scale now.
                    self.register_scale("scale_weight", mod_extra_config.scale.params["weight"][0], ScaleFormat.CONST)

            self.is_dynamic_quantization = isinstance(self.quant_input, QuantDynamicInput)
            self.quant_input_func = self.quant_input_and_get_scale_dynamic if self.is_dynamic_quantization else self.quant_input_and_get_scale_static
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            init_mixture_of_experts_linears(self)

    def quant_input_and_get_scale_dynamic(self, input):
        return self.quant_input(input)

    def quant_input_and_get_scale_static(self, input):
        return self.quant_input(input), self.scale_input

    def run_linear_quant(self, input, bias):
        qinput, scale_input = self.quant_input_func(input)
        y = self.matmul_fp8(qinput,
                            self.weight,
                            out_dtype=self._mod_extra_config.config_params["hp_dtype"],
                            scale_input_inv=scale_input,
                            scale_other_inv=self.scale_weight)
        output = y + bias if (bias is not None) else y
        return output

    def run_linear_qdq(self, input, bias):
        if self.fake_quant:
            qweight = self.qdq_weights(self.weight, )
        else:
            qweight = self.dequant_weights(self.weight, )
        qinput = self.quant_input(input)
        y = torch.matmul(qinput, qweight)
        output = y + bias if (bias is not None) else y
        return output

    def forward_qdq(self, input):
        return self.run_linear_qdq(input, self.bias)

    def forward_quant(self, input):
        return self.run_linear_quant(input, self.bias)

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight") if not self.use_qdq else get_current_repr(self),
        )


class PatchedLinear(PatchedLinearBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.init_linear(mod_extra_config)

    def forward_measure(self, input):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input)
        measure_output((output,), self._mod_extra_config.outputs)
        return output
    
    
class PatchedParallelLMHead(PatchedLinearBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        # ParallelLMHead inherits from VocabParallelEmbedding (nn.module) which has a member called
        # "quant_method" of type UnquantizedEmbeddingMethod that inherits from QuantizeMethodBase
        # (both are not nn.module) and implement an "apply" method by using torch.nn.functional.linear
        # (Apply the weights in layer to the input tensor.)
        # ParallelLMHead's forward method should not be called because LMHead's weights should be used
        # in the sampler. (The forward itself throws RuntimeError exception)
        # So in order to quantize that quant_method we patch only the "apply" method.
        self.init_linear(mod_extra_config)
        self.orig_linear_quant_apply = self.orig_mod.quant_method.apply
        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            if self.use_qdq or self.fake_quant:
                self.quant_method.apply = self.apply_qdq
            else:
                self.quant_method.apply = self.apply_quant
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.quant_method.apply = self.apply_measure

    def apply_quant(self, layer, x, bias):
        return self.run_linear_quant(x, bias)

    def apply_qdq(self, layer, x, bias):
        return self.run_linear_qdq(x, bias)

    def apply_measure(self, layer, x, bias):
        measure_input((x,), observer=self._mod_extra_config.inputs)
        output = self.orig_linear_quant_apply(layer, x, bias)
        measure_output((output,), self._mod_extra_config.outputs)
        return output


class PatchedReplicatedLinear(PatchedLinearBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.init_linear(mod_extra_config)

    def forward_qdq(self, input):
        bias = self.bias if not self.skip_bias_add else None
        output = self.run_linear_qdq(input, bias)
        output_bias = self.bias if self.skip_bias_add else None
        return output, output_bias

    def forward_quant(self, input):
        bias = self.bias if not self.skip_bias_add else None
        output = self.run_linear_quant(input, bias)
        output_bias = self.bias if self.skip_bias_add else None
        return output, output_bias

    def forward_measure(self, input):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output, output_bias = self.orig_mod(input)
        measure_output((output,), self._mod_extra_config.outputs)
        return output, output_bias


class PatchedLinearAllReduce(PatchedLinearBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.init_linear(mod_extra_config)
        self.scoped_version = mod.__class__.__name__ == "ScopedLinearAllReduce"

    def forward_qdq(self, input):
        # pre_all_reduce
        output = self.run_linear_qdq(input, None)

        if not self.scoped_version:
            self.all_reduce(output)
            output = self.post_all_reduce(output)
        return output

    def forward_quant(self, input):
        # pre_all_reduce
        output = self.run_linear_quant(input, None)
        # TODO: enable dynamic quantization for output
        dqoutput = self.dequant_output(output)
        if not self.scoped_version:
            self.all_reduce(dqoutput)
            dqoutput = self.post_all_reduce(dqoutput)
        return dqoutput

    def forward_measure(self, input):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = torch.matmul(input, self.weight.transpose(-1, -2))
        measure_output((output,), self._mod_extra_config.outputs)
        # in scoped version all reduce is being called outside of the layer
        if not self.scoped_version:
            self.all_reduce(output)
            output = self.post_all_reduce(output)
        return output

    def all_reduce(self, input):
        if self.mp_group is not None:
            from deepspeed import comm as dist

            dist.inference_all_reduce(input, group=self.mp_group)

    def post_all_reduce(self, input):
        output = input + self.bias if (self.bias is not None) else input
        return output


class PatchedRowParallelLinear(PatchedLinearBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        from .._core.external_func_impl import get_external_row_parallel_collective_func
        self.row_parallel_collective_func = get_external_row_parallel_collective_func()
        # TODO [SW-224403]: Enable dynamic quantization in row parallel allreduce
        allreduce_quantization_enable = get_hqt_config(mod).cfg["row_parallel_linear_allreduce_quantization"]
        if self.quantization_mode in (QuantMode.MEASURE, QuantMode.SHAPE):
            self.forward = self.forward_measure_reduce if self.reduce_results and self.tp_size > 1 else self.forward_measure_no_reduce

        elif self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            if self.fake_quant or self.use_qdq:
                self.forward = self.forward_qdq
            else:
                if self.reduce_results and self.tp_size > 1:
                    if allreduce_quantization_enable:
                        self.forward = self.forward_quant_reduce_in_lp
                    else:
                        self.forward = self.forward_quant_reduce_in_hp
                else:
                    self.forward = self.forward_quant_no_reduce
        self.init_linear(mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            if allreduce_quantization_enable:
                self.dequant_scatter_output = self._mod_extra_config.outputs[0]
                self.quant_scatter_input = self._mod_extra_config.outputs[1]
                self.quant_gather_input = self._mod_extra_config.outputs[2]
                self.dequant_gather_output = self._mod_extra_config.outputs[3]
        from torch import distributed as dist
        self.world_size = dist.get_world_size()

    def resolve_input(self, input_):
        """
        this code is copied from vllm RowParallelLinear forward method
        """
        if self.input_is_parallel:
            input_parallel = input_
        else:
            tp_rank = get_tensor_model_parallel_rank()
            splitted_input = split_tensor_along_last_dim(
                input_, num_partitions=self.tp_size)
            input_parallel = splitted_input[tp_rank].contiguous()
        return input_parallel

    def forward_qdq(self, input):
        # TODO: [SW-208441] Support all_reduce_fp8 in forward_qdq in PatchedRowParallelLinear
        resolved_input = self.resolve_input(input)
        output = self.run_linear_qdq(resolved_input, None)

        if self.reduce_results:
            output = self.row_parallel_collective_func(output)
        return self.bias_add(output)

    def lp_matmul_hp(self, input):
        """
        Perfoms a matmul with lp inputs and returns output in hp
        """
        resolved_input = self.resolve_input(input)
        return self.run_linear_quant(resolved_input, None)

    def forward_quant_no_reduce(self, input):
        matmul_output_hp = self.lp_matmul_hp(input)
        return self.bias_add(matmul_output_hp)

    def forward_quant_reduce_in_lp(self, input):
        matmul_output_hp = self.lp_matmul_hp(input)
        # performing all_reduce in fp8 should be done only at decode phase
        # decode phase is indicated by input.shape[1] == 1
        if input.shape[1] == 1:
            allreduce_output_hp = self.quant_all_reduce_sum(matmul_output_hp)
        else:
            allreduce_output_hp = self.row_parallel_collective_func(matmul_output_hp)
        return self.bias_add(allreduce_output_hp)

    def forward_quant_reduce_in_hp(self, input):
        matmul_output_hp = self.lp_matmul_hp(input)
        all_reduce_output_hp = self.row_parallel_collective_func(matmul_output_hp)
        return self.bias_add(all_reduce_output_hp)

    def measure_input_and_matmul(self, input):
        resolved_input = self.resolve_input(input)
        measure_input((resolved_input,), observer=self._mod_extra_config.inputs)
        return self.orig_mod.quant_method.apply(self.orig_mod, resolved_input)

    def forward_measure_no_reduce(self, input):
        output = self.measure_input_and_matmul(input)
        temp = torch.empty(1)
        measure_output((output, temp,), self._mod_extra_config.outputs)
        return self.bias_add(output)

    def forward_measure_reduce(self, input):
        from torch import distributed as dist
        output = self.measure_input_and_matmul(input)
        max_output = output.clone()
        dist.all_reduce(max_output, op=dist.ReduceOp.MAX)
        all_reduce_output = self.row_parallel_collective_func(output)
        measure_output((max_output, all_reduce_output,), self._mod_extra_config.outputs)
        return self.bias_add(all_reduce_output)

    def scatter(self, tensor):
        from torch import distributed as dist
        buffer = torch.empty_like(tensor)
        # Perform all_to_all_single communication
        dist.all_to_all_single(buffer, tensor)
        # Reshape output_tensor to (world_size, -1)
        return buffer.view(self.world_size, -1)

    def gather(self, tensor, reduced_chunk, original_shape):
        from torch import distributed as dist
        # Gather the reduced chunks from all processes
        dist.all_gather_into_tensor(tensor, reduced_chunk)
        # Reshape back to the original tensor shape
        return tensor.view(original_shape)

    def quant_all_reduce_sum(self, scatter_input_hp):
        scatter_input_lp = self.quant_scatter_input(scatter_input_hp)
        original_shape = scatter_input_lp.shape
        buffer_lp = self.scatter(scatter_input_lp)

        # Dequant tensor as sum preformed in bf16
        buffer_hp = self.dequant_scatter_output(buffer_lp)
        # Sum the received chunks (local reduction)
        sum_hp = buffer_hp.sum(dim=0)  # Shape: [-1]
        # Quant tensor as all_gather preformed in fp8
        gather_input_lp = self.quant_gather_input(sum_hp)

        gather_output_lp = self.gather(scatter_input_lp, gather_input_lp, original_shape)
        gather_output_hp = self.dequant_gather_output(gather_output_lp)

        return gather_output_hp

    def bias_add(self, output):
        assert (
            self.reduce_results or (not self.bias) or self.skip_bias_add
        ), "When not reduce the results, adding bias to the results can lead to incorrect results"
        if not self.skip_bias_add:
            output = output + self.bias if self.bias is not None else output
            output_bias = None
        else:
            output_bias = self.bias
        return output, output_bias


class PatchedColumnParallelLinear(PatchedLinearBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.init_linear(mod_extra_config)
        from .._core.external_func_impl import get_external_column_parallel_collective_func
        self.column_parallel_collective_func = get_external_column_parallel_collective_func()

    def forward_qdq(self, input):
        output = self.run_linear_qdq(input, None)
        output, output_bias = self.add_bias(output)
        if self.gather_output:
            output = self.column_parallel_collective_func(output)
        return output, output_bias

    def forward_quant(self, input):
        output = self.run_linear_quant(input, None)
        # TODO: enable dynamic quantization for output
        dqoutput = self.dequant_output(output)
        dqoutput, dqoutput_bias = self.add_bias(dqoutput)
        if self.gather_output:
            dqoutput = self.column_parallel_collective_func(dqoutput)
        return dqoutput, dqoutput_bias

    def forward_measure(self, input):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod.quant_method.apply(self.orig_mod, input)
        measure_output((output,), self._mod_extra_config.outputs)
        output, output_bias = self.add_bias(output)
        if self.gather_output:
            output = self.column_parallel_collective_func(output)
        return output, output_bias

    def add_bias(self, output):
        if not self.skip_bias_add:
            output = output + self.bias if self.bias is not None else output
            output_bias = None
        else:
            output_bias = self.bias
        return output, output_bias


class PatchedLmHeadLinearAllreduce(PatchedLinearBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.init_linear(mod_extra_config)

    def forward_qdq(self, input):
        assert (
            input.shape[-1] % self.world_size == 0
        ), "Please ensure that self.world_size is divisible by input.shape[-1]"
        input_shard = input.shape[-1] // self.world_size
        splittedInput = input[:, :, self.rank * input_shard: (self.rank + 1) * input_shard]
        output = self.run_linear_qdq(splittedInput, None)

        if self.mp_group is not None:
            from deepspeed import comm as dist

            dist.inference_all_reduce(output, group=self.mp_group)
        if self.bias is not None:
            output += self.bias
        return output

    def forward_quant(self, input):
        from deepspeed.module_inject.tp_shard import get_shard_size, get_shard_size_list
        input_shard_size = get_shard_size(input.shape[-1], self.world_size, "lm_head")
        input_shard_offset = sum(get_shard_size_list(input.shape[-1], self.world_size, "lm_head")[0:self.rank])
        splittedInput = input[:, :, input_shard_offset:input_shard_offset + input_shard_size]
        output = self.run_linear_quant(splittedInput, None)
        # TODO: enable dynamic quantization for output
        dqoutput = self.dequant_output(output)

        if self.mp_group is not None:
            from deepspeed import comm as dist

            dist.inference_all_reduce(dqoutput, group=self.mp_group)
        if self.bias is not None:
            dqoutput += self.bias
        return dqoutput

    def forward_measure(self, input):
        from deepspeed.module_inject.tp_shard import get_shard_size, get_shard_size_list
        input_shard_size = get_shard_size(input.shape[-1], self.world_size, "lm_head")
        input_shard_offset = sum(get_shard_size_list(input.shape[-1], self.world_size, "lm_head")[0:self.rank])
        splittedInput = input[:, :, input_shard_offset:input_shard_offset + input_shard_size]
        measure_input((splittedInput,), observer=self._mod_extra_config.inputs)
        output = torch.matmul(splittedInput, self.weight.t())
        measure_output((output,), self._mod_extra_config.outputs)

        if self.mp_group is not None:
            from deepspeed import comm as dist

            dist.inference_all_reduce(output, group=self.mp_group)
        if self.bias is not None:
            output += self.bias
        return output


class PatchedEmbeddingBag(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            if self.use_qdq:
                self.dequant_weights = self._mod_extra_config.params["weight"][1]
                if isinstance(mod_extra_config.scale.params["weight"], (torch.Tensor, float)):
                    self.register_scale("scale_weight", mod_extra_config.scale.params["weight"], self.scale_format)
                elif isinstance(mod_extra_config.scale.params["weight"], dict):
                    # PCQ weight is calculated with actual weight [0] and ones [1]
                    # only ScaleFormat.CONST is supported for per-channel scale now.
                    self.register_scale("scale_weight", mod_extra_config.scale.params["weight"][0], ScaleFormat.CONST)
            else:
                raise ValueError("EmbeddingBag is only supported QDQ mode now!")

    def forward_qdq(self, input, offsets, *args, **kwargs):
        qweight = self.dequant_weights(self.weight, )

        return torch.nn.functional.embedding_bag(
            input=input,
            offsets=offsets,
            weight=qweight,
            max_norm=self.max_norm,
            norm_type=self.norm_type,
            scale_grad_by_freq=self.scale_grad_by_freq,
            mode=self.mode,
            sparse=self.sparse,
            include_last_offset=self.include_last_offset,
            padding_idx=self.padding_idx,
            *args,
            **kwargs,
        )

    def forward_measure(self, input, *args, **kwargs):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input, *args, **kwargs)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_weight"),
        )


# patched vllm FusedMoE module removing the bf16 weights of all experts
# measure and quant of the weights is done per expert using PatchedMoeMatmul
# therefore it is configured: ModuleInfo.should_measure_and_quant = False
class PatchedMixtralMoE(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        # remove the MoE weights that are quanted by PatchedMoeMatmul
        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            delattr(mod, "w13_weight")
            delattr(mod, "w2_weight")
            setattr(mod, "w13_weight", None)
            setattr(mod, "w2_weight", None)
            setattr(self, "w13_weight", None)
            setattr(self, "w2_weight", None)
        self.forward = self.forward_orig

    # copied from https://github.com/HabanaAI/vllm-fork/blob/93b8bad8478451349d0c76b3116d3ad863a3b48e/vllm/model_executor/layers/fused_moe/layer.py#L1429
    def maybe_all_reduce_tensor_model_parallel(
            self, final_hidden_states: torch.Tensor):
        """
        The pplx combine kernel reduces across GPU ranks by default.
        """
        if (self.orig_mod.use_pplx_kernels or self.orig_mod.use_deepep_ht_kernels
                or self.orig_mod.use_deepep_ll_kernels):
            return final_hidden_states
        else:
            from .._core.external_func_impl import get_external_row_parallel_collective_func
            tensor_model_parallel_all_reduce = get_external_row_parallel_collective_func()
            return tensor_model_parallel_all_reduce(final_hidden_states)

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self),
        )


# This patched module is called by the vllm-mixtral FusedMoE layer
# we wrap each expert weight with this module since FusedMoE has a single tensor for all experts weights
# this way we can calculate scales per expert and achive better accuracy
class PatchedMoeMatmul(PatchedLinearBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.init_linear(mod_extra_config)
        if (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            measure_input((torch.tensor(0),), observer=self._mod_extra_config.inputs)
        else:
            self.weight = torch.nn.Parameter(self.weight.squeeze(), requires_grad=False)

    def forward_qdq(self, input, *args, **kwargs):
        return self.run_linear_qdq(input, None)

    def forward_quant(self, input, *args, **kwargs):
        return self.run_linear_quant(input, None)

    def forward_measure(self, input, *args, **kwargs):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input, *args, **kwargs)
        measure_output((output,), self._mod_extra_config.outputs)
        return output


class PatchedGaudiMixtralSparseMoeBlock(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.forward = self.forward_orig
        self.ep_size = mod.ep_size
        self.experts_min = mod.experts_min
        self.experts_max = mod.experts_max
        self.experts_range = mod.experts_range
        self.num_experts = mod.num_experts

        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            self.dynamic_moe_op = get_quantized_func_wrapper(OP_TYPE.DYNAMIC_MOE, self.scale_format)
            self.quant_input = self._mod_extra_config.inputs[0]
            self.register_scale("scale_input", mod_extra_config.scale.inputs[0], self.scale_format)
            self.register_scale(
                "scale_intermediate",
                [mod_extra_config.scale.inputs[x] for x in range(self.experts_min + 1, self.experts_max + 2)],
                self.scale_format,
            )
            mod.call_dynamic_moe_op = get_call_wrapper(self, "call_dynamic_moe_quant_op")
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            mod.call_dynamic_moe_op = self.call_dynamic_moe_measure_op

    def call_dynamic_moe_quant_op(self,
                                  hidden_states,
                                  expert_routing_table,
                                  router_weights,
                                  permuted_weights=False,
                                  activation="silu"):
        w1_list = [self.experts[i].w1.weight for i in self.experts_range]
        w2_list = [self.experts[i].w2.weight for i in self.experts_range]
        w3_list = [self.experts[i].w3.weight for i in self.experts_range]
        scale_w1 = [self.experts[i].w1.scale_weight for i in self.experts_range]
        scale_w2 = [self.experts[i].w2.scale_weight for i in self.experts_range]
        scale_w3 = [self.experts[i].w3.scale_weight for i in self.experts_range]
        qinput = self.quant_input(hidden_states)
        output = self.dynamic_moe_op(
            hidden_states=qinput,
            expert_routing_table=expert_routing_table,
            router_weights=router_weights,
            w1=w1_list,
            w3=w2_list,
            w2=w3_list,
            d_scale_w1=scale_w1,
            d_scale_w2=scale_w3,
            d_scale_w3=scale_w2,
            d_scale_hidden_states=self.scale_input,
            d_scale_intermediate_hidden_states=self.scale_intermediate,
            permuted_weights=False,
            activation=activation,
            experts_min=self.experts_min,
            experts_max=self.experts_max,
        )
        return output

    def call_dynamic_moe_measure_op(self,
                                    hidden_states,
                                    expert_routing_table,
                                    router_weights,
                                    permuted_weights=True,
                                    activation="silu"):
        w1_list = [self.experts[i].w1.weight for i in self.experts_range]
        w2_list = [self.experts[i].w2.weight for i in self.experts_range]
        w3_list = [self.experts[i].w3.weight for i in self.experts_range]
        measure_input((hidden_states,), observer=self._mod_extra_config.inputs)

        output, intermidiate_amax = torch.ops.hpu.mixture_of_experts.fp8_measurement(
            hidden_states=hidden_states,
            expert_routing_table=expert_routing_table,
            router_weights=router_weights,
            w1=w1_list,
            w3=w2_list,
            w2=w3_list,
            permuted_weights=permuted_weights,
            activation=activation,
            experts_min=self.experts_min,
            experts_max=self.experts_max,
            measurement_mode=True,
        )

        amax = []
        for i in range(self.num_experts):
            if i in self.experts_range:
                amax.append(intermidiate_amax[i - self.experts_min])
            else:
                amax.append(torch.tensor(0, device="hpu", dtype=intermidiate_amax[0].dtype))

        output_measure_list = [output] + amax

        measure_output(output_measure_list, self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        member_names = ["scale_input"]
        for x in range(1, self.num_experts + 1):
            member_names.append("scale_intermediate[" + str(x) + "]")
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, *member_names),
        )


class PatchedGaudiDeepseekV3MoE(PatchedGaudiMixtralSparseMoeBlock):
    """The patched module for the Optimum Habana DeepseekV3MoE BF16 and FP8 models.
    
    This MoE module differs from the Mixtral MoE by chunking the Gaudi mixture of 
    experts op for models like DeepSeek with large number of experts. The
    size of the chunks is set using SLICE_MAX_EXPERT environment variable.

    For FP8 models, we need to requantize weights as follows:
    - At measurement stage, we dequantize the FP8 weights to BF16.
    - At quantization stage, we use the same `forward_quant` method for both BF16 and FP.
    """
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)

        # Check if FP8 model
        self._is_fp8 = False
        if hasattr(mod.config, "quantization_config"):
            if hasattr(mod.config.quantization_config, "quant_method"):
                self._is_fp8 = (mod.config.quantization_config.quant_method == 'fp8')
                
    def call_dynamic_moe_quant_op(self,
                                  hidden_states,
                                  expert_routing_table,
                                  router_weights,
                                  permuted_weights=False,
                                  activation="silu"):
        
        gate_proj_list = [self.experts[i].gate_proj.weight for i in self.experts_range]
        down_proj_list = [self.experts[i].down_proj.weight for i in self.experts_range]
        up_proj_list = [self.experts[i].up_proj.weight for i in self.experts_range]
        scale_gate_proj = [self.experts[i].gate_proj.scale_weight for i in self.experts_range]
        scale_down_proj = [self.experts[i].down_proj.scale_weight for i in self.experts_range]
        scale_up_proj = [self.experts[i].up_proj.scale_weight for i in self.experts_range]
        qinput = self.quant_input(hidden_states)
        return self.dynamic_moe_op(
            hidden_states=qinput,
            expert_routing_table=expert_routing_table,
            router_weights=router_weights,
            w1=gate_proj_list,
            w2=up_proj_list,
            w3=down_proj_list,
            d_scale_w1=scale_gate_proj,
            d_scale_w2=scale_up_proj,
            d_scale_w3=scale_down_proj,
            d_scale_hidden_states=self.scale_input,
            d_scale_intermediate_hidden_states=self.scale_intermediate,
            permuted_weights=False,
            activation=activation,
            experts_min=self.experts_min,
            experts_max=self.experts_max,
        )
        
    def call_dynamic_moe_measure_op(self,
                                    hidden_states,
                                    expert_routing_table,
                                    router_weights,
                                    permuted_weights=True,
                                    activation="silu"):
        if self._is_fp8:
            gate_proj_list = [self.experts[i].gate_proj.get_dequant_weight() for i in self.experts_range]
            down_proj_list = [self.experts[i].down_proj.get_dequant_weight() for i in self.experts_range]
            up_proj_list = [self.experts[i].up_proj.get_dequant_weight() for i in self.experts_range]
        else:
            gate_proj_list = [self.experts[i].gate_proj.weight for i in self.experts_range]
            down_proj_list = [self.experts[i].down_proj.weight for i in self.experts_range]
            up_proj_list = [self.experts[i].up_proj.weight for i in self.experts_range]
        
        measure_input((hidden_states,), observer=self._mod_extra_config.inputs)
        output, intermediate_amax = torch.ops.hpu.mixture_of_experts.fp8_measurement(
            hidden_states=hidden_states,
            expert_routing_table=expert_routing_table,
            router_weights=router_weights,
            w1=gate_proj_list,
            w2=up_proj_list,
            w3=down_proj_list,
            permuted_weights=permuted_weights,
            activation=activation,
            experts_min=self.experts_min,
            experts_max=self.experts_max,
            measurement_mode=True,
        )

        amax = []
        for i in range(self.num_experts):
            if i in self.experts_range:
                amax.append(intermediate_amax[i-self.experts_min])
            else:
                amax.append(torch.tensor(0, device="hpu", dtype=intermediate_amax[0].dtype))

        output_measure_list = [output] + amax

        measure_output(output_measure_list, self._mod_extra_config.outputs)
        return output

    
class PatchedVllmMixtureOfExpertsOp(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        # Get the `experts_min` and `experts_max` from the original module if they exist
        self.experts_min = self.orig_mod.experts_min if hasattr(self.orig_mod, "experts_min") else 0
        self.experts_max = self.orig_mod.experts_max if hasattr(self.orig_mod, "experts_max") else 7
        self.experts_used = self.local_num_experts if hasattr(self.orig_mod, "local_num_experts") else self.num_experts
        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:

            self.quant_input = self._mod_extra_config.inputs[0]
            self.register_scale("scale_input", mod_extra_config.scale.inputs[0], self.scale_format)
            self.register_scale(
                "scale_intermediate",
                [mod_extra_config.scale.inputs[x] for x in range(1, self.experts_used + 1)],
                self.scale_format,
            )
            self.is_dynamic_quantization = isinstance(self.quant_input, QuantDynamicInput)
            self.dynamic_moe_op = get_quantized_func_wrapper(
                OP_TYPE.DYNAMIC_MOE_FUSED_WEIGHTS, scale_format=self.scale_format, is_dynamic=self.is_dynamic_quantization
            )
            if self.is_dynamic_quantization:
                self.forward = self.forward_dynamic_quant

    def _get_extra_kwargs(self, tokens_num: int):
        kwargs = {}
        if hasattr(self.orig_mod, "_get_extra_kwargs"):
            kwargs = self.orig_mod._get_extra_kwargs(tokens_num)
        return kwargs

    def forward_quant(self,
                      hidden_states,
                      expert_routing_table,
                      router_weights,
                      permuted_weights=True,
                      activation="silu"):
        tokens_num, hidden_dim = hidden_states.shape
        extra_kwargs = self._get_extra_kwargs(tokens_num)
        experts_range = range(self.experts_used)
        w1_list = [self.w13_list[i].weight for i in experts_range]
        w2_list = [self.w2_list[i].weight for i in experts_range]
        scale_w1 = [self.w13_list[i].scale_weight for i in experts_range]
        scale_w2 = [self.w2_list[i].scale_weight for i in experts_range]
        qinput = self.quant_input(hidden_states)
        output = self.dynamic_moe_op(
            hidden_states=qinput,
            expert_routing_table=expert_routing_table,
            router_weights=router_weights,
            w12=w1_list,
            w3=w2_list,
            d_scale_w12=scale_w1,
            d_scale_w3=scale_w2,
            d_scale_hidden_states=self.scale_input,
            d_scale_intermediate_hidden_states=self.scale_intermediate,
            permuted_weights=False,
            activation=activation,
            experts_min=self.experts_min,
            experts_max=self.experts_max,
            **extra_kwargs,
        )
        return output

    def forward_dynamic_quant(
        self, hidden_states, expert_routing_table, router_weights, permuted_weights=True, layer=None, activation="silu"
    ):
        # This is the dynamic version of the forward_quant method.
        # Compared to the `forward_quant` method, the main differences are:
        #   1) The `quant_input` is of type `QuantDynamicInput`.
        #   2) There is no need to pass the `d_scale_intermediate_hidden_states` to the dynamic moe op.
        experts_range = range(self.num_experts)
        w1_list = [self.w13_list[i].weight for i in experts_range]
        w2_list = [self.w2_list[i].weight for i in experts_range]
        scale_w1 = [self.w13_list[i].scale_weight for i in experts_range]
        scale_w2 = [self.w2_list[i].scale_weight for i in experts_range]
        qinput_fp8, input_scale = self.quant_input(hidden_states)
        output = self.dynamic_moe_op(
            hidden_states=qinput_fp8,
            expert_routing_table=expert_routing_table,
            router_weights=router_weights,
            w12=w1_list,
            w3=w2_list,
            d_scale_w12=scale_w1,
            d_scale_w3=scale_w2,
            d_scale_hidden_states=input_scale,
            permuted_weights=False,
            activation=activation,
            experts_min=self.experts_min,
            experts_max=self.experts_max
        )
        return output

    def forward_measure(self,
                        hidden_states,
                        expert_routing_table,
                        router_weights,
                        permuted_weights=True,
                        activation="silu"):
        experts_range = range(self.experts_used)
        w1_list = [self.w13_list[i].weight.squeeze() for i in experts_range]
        w2_list = [self.w2_list[i].weight.squeeze() for i in experts_range]
        measure_input((hidden_states,), observer=self._mod_extra_config.inputs)
        output, intermidiate_amax = torch.ops.hpu.mixture_of_experts.fp8_measurement_fused_weights(
            hidden_states=hidden_states,
            expert_routing_table=expert_routing_table,
            router_weights=router_weights,
            w12=w1_list,
            w3=w2_list,
            permuted_weights=permuted_weights,
            activation=activation,
            experts_min=self.experts_min,
            experts_max=self.experts_max,
            measurement_mode=True,
        )
        output_measure_list = [output]
        for i in range(self.num_experts):
            output_measure_list.append(intermidiate_amax[i])
        measure_output(output_measure_list, self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        member_names = ["scale_input", "scale_intermediate"]
        # for x in range(1, self.num_experts+1):
        #     member_names.append("scale_intermediate["+str(x)+"]")
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, *member_names),
        )


class PatchedVllmMixtureOfExpertsOpFP8(PatchedVllmMixtureOfExpertsOp):
    """The patched module for the VLLMMixtureOfExpertsOp module with FP8 weights.

    There are some models like Deepseek R1/V3 with FP8 weights, we need to requantize it.

    The main difference between this module and the PatchedVllmMixtureOfExpertsOp is that the weights are FP8.
    - At measurement stage, we dequantize the weights to BF16.
    - At quantization stage, we use the same `forward_quant` method as the PatchedVllmMixtureOfExpertsOp.
    """

    def forward_measure(
            self,
            x,
            topk_ids,
            topk_weights,
            permuted_weights=True,
            activation="silu"):
        hidden_states = x
        measure_input((hidden_states,), observer=self._mod_extra_config.inputs)
        min_expert = self.experts_min
        max_expert = self.experts_max
        w13_list_slice = []
        w2_list_slice = []
        for j in range(self.num_experts):
            w13_list_slice.append(self.w13_list[j].get_dequant_weight())
            w2_list_slice.append(self.w2_list[j].get_dequant_weight())

        output, intermidiate_amax = torch.ops.hpu.mixture_of_experts.fp8_measurement_fused_weights(
            hidden_states=x,
            expert_routing_table=topk_ids.to(torch.int64),
            router_weights=topk_weights.to(x.dtype),
            w12=w13_list_slice,
            w3=w2_list_slice,
            permuted_weights=permuted_weights,
            activation=activation,
            experts_min=min_expert,
            experts_max=max_expert,
            measurement_mode=True,
        )
        output_measure_list = [output]
        for i in range(self.experts_used):
            output_measure_list.append(intermidiate_amax[i])
        measure_output(output_measure_list, self._mod_extra_config.outputs)
        return output


class PatchedMoeFP8Matmul(PatchedMoeMatmul):
    """The patched module for the MoeMatmul module with FP8 weights."""

    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.get_dequant_weight = self.orig_mod.get_dequant_weight


class PatchedKVCache(PatchedModuleBase):
    # Module to patch KVCache module from llama model
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        kwargs["func_names"] = ("forward", "get_shape")
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.org_allocate = mod.allocate
        self.org_update = mod.update
        self.forward = self.forward_orig

        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.dequant_output = self._mod_extra_config.outputs[0]
            if self.use_qdq:
                self.qdq_input = self._mod_extra_config.inputs[1]
                self.update = self.update_qdq
                mod.update = self.update_qdq
            else:
                self.update = self.update_quant
                mod.update = self.update_quant
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.update = self.update_measure
            mod.update = self.update_measure

    # overwrite allocate function of original module to force allocation in fp8
    def allocate(self, inp_seq_len, dtype, device, shape):
        dtype = torch.float8_e4m3fn if (self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]) else dtype
        return self.org_allocate(inp_seq_len, dtype, device, shape)

    # overwrite update function of original module to force quant and dequant of cache input and output
    def update_qdq(self, prev, cur, dim, idx, inp_seq_len):
        """
         Explanation:  If we want to optimize index_copy so it would run in fp8 instead of bf16
                       we need the tensors to be in fp8 before calling index_copy.
                       Also the `prev` and `curr` tensors need to be of the same dtype - and quanting them both
                       from bf16 is no help, best we can do is have prev be initialized an fp8 tensor from the start.
                       Since the initilization of `prev` is done in OHF (and that is not implemented yet) we
                       currently need to support both options until the implementation in OHF is done, then
                       can we remove the support for the bf16 `prev` option (the else here).
        """
        if prev.dtype == torch.float8_e4m3fn:
            qcurr = self.quant_input(cur)
            qoutput = self.org_update(prev, qcurr, dim, idx, inp_seq_len)
            output = self.dequant_output(qoutput)
        # TODO: remove the `else` part once the lp_dtype is implemented in OHF
        else:
            curr = self.qdq_input(cur)
            output = self.org_update(prev, curr, dim, idx, inp_seq_len)
        return output

    # overwrite update function of original module to force quant and dequant of cache input and output
    def update_quant(self, prev, cur, dim, idx, inp_seq_len):
        qinput = self.quant_input(cur)
        output = self.org_update(prev, qinput, dim, idx, inp_seq_len)
        if output.dtype == torch.float8_e4m3fn:
            return self.dequant_output(output)
        else:
            return output

    # overwrite update function of original module to force quant and dequant of cache input and output
    def update_measure(self, prev, cur, dim, idx, inp_seq_len):
        measure_input((cur,), self._mod_extra_config.inputs)
        output = self.org_update(prev, cur, dim, idx, inp_seq_len)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self),
        )


class PatchedVLLMKVCache(PatchedModuleBase):
    # Module to patch VLLMKVCache module from llama model
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.dequant_output = self._mod_extra_config.outputs[0]
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.fetch_from_cache = mod.fetch_from_cache

    def forward_qdq(self, input, *args, **kwargs):
        qinput = self.quant_input(input)
        output_cache = self.orig_mod(qinput, *args, **kwargs)
        return output_cache

    def forward_quant(self, input, cache, *args, **kwargs):
        if input is not None:
            qinput = self.quant_input(input)
            output_cache = self.orig_mod(qinput, cache, *args, **kwargs)
        else:
            # In cross-attention during decode stage kv cache isn't updated
            # so input is None and we don't store it
            output_cache = cache
        return self.dequant_output(output_cache)

    def forward_measure(self, input, cache, *args, **kwargs):
        # In cross-attention during decode stage kv cache isn't updated
        # so input is None and we don't measure it
        if input is None:
            return cache
        measure_input((input, ), self._mod_extra_config.inputs)
        output_cache = self.orig_mod(input, cache, *args, **kwargs)
        measure_output((output_cache, ), self._mod_extra_config.outputs)
        return output_cache

    def fetch_from_cache(self, cache, blocks):
        if cache.dtype != self.lp_dtype:
            quant_cache = self.quant_input(cache)
        else:
            quant_cache = cache
        output_cache = self.orig_mod.fetch_from_cache(quant_cache, blocks)
        return self.dequant_output(output_cache)

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self),
        )


def init_conv(instance, mod_extra_config):
    if instance.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
        instance.quant_input = instance._mod_extra_config.inputs[0]
        if instance.use_qdq:
            instance.dequant_weights = mod_extra_config.params["weight"][1]
        else:
            instance.register_scale("scale_input", mod_extra_config.scale.inputs[0], instance.scale_format)
            instance.register_scale("scale_weight", mod_extra_config.scale.params["weight"], instance.scale_format)
            instance.conv2d_fp8 = get_quantized_func_wrapper(OP_TYPE.CONV, instance.scale_format)


class PatchedConv2d(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        init_conv(self, mod_extra_config)

    def forward_qdq(self, input):
        qweight = self.dequant_weights(self.weight, )
        qinput = self.quant_input(input)
        output = torch.nn.functional.conv2d(
            qinput,
            qweight,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
        )
        return output

    def forward_quant(self, input):
        qinput = self.quant_input(input)
        output = self.conv2d_fp8(qinput,
                                 self.weight,
                                 self.bias,
                                 self.stride,
                                 self.padding,
                                 self.dilation,
                                 self.groups,
                                 out_dtype=self._mod_extra_config.config_params["hp_dtype"],
                                 scale_input_inv=self.scale_input,
                                 scale_other_inv=self.scale_weight)
        return output

    def forward_measure(self, input):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight") if not self.use_qdq else "",
        )


class PatchedSoftmaxBase(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            self.dequant_output = self._mod_extra_config.outputs[0]
            if self.use_qdq:
                self.quant_input = qdq(nn.Parameter(torch.Tensor([1.0])), self.lp_dtype, self.hp_dtype)
            else:
                # input scale is 1 assuming the input to SM is descaled because we are using HW supported scales
                self.register_scale("scale_input", torch.Tensor([1.0]), self.scale_format)
                self.register_scale("scale_output", torch.Tensor(
                    [1 / mod_extra_config.scale.outputs[0]]), self.scale_format)
            self.softmax_kernel_op = get_quantized_func_wrapper(
                self.get_op_type(), self.scale_format)

    @staticmethod
    @abstractmethod
    def get_op_type():
        """Forward function with Q-DQ mode for quantization stage."""
        raise NotImplementedError("get_op_type is not implemented")

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_output") if not self.use_qdq else "",
        )

class PatchedSoftmax(PatchedSoftmaxBase):

    @staticmethod
    def get_op_type():
        return OP_TYPE.SOFTMAX

    def forward_qdq(self, x, dim=None, invAttnHead=None):
        x = self.quant_input(x)
        output = torch.softmax(x, dim)
        return output

    def forward_quant(self, x, dim=None, invAttnHead=None):
        output = self.softmax_kernel_op(x, dim, self.scale_input, self.scale_output, invAttnHead)
        return self.dequant_output(output)

    def forward_measure(self, x, dim=None, invAttnHead=None):
        measure_input((x,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(x, dim)
        measure_output((output,), self._mod_extra_config.outputs)
        return output


class PatchedBlockSoftmaxConstMax(PatchedSoftmaxBase):

    @staticmethod
    def get_op_type():
        return OP_TYPE.BLOCK_SOFTMAX_CONST_MAX

    def forward_qdq(self, attn, block_bias, block_groups, batch_size, global_max):
        q_attn = self.quant_input(attn)
        output = self.softmax_kernel_op(q_attn, block_bias, block_groups, batch_size, global_max)
        return output

    def forward_quant(self, attn, block_bias, block_groups, batch_size, global_max):
        output_fp8 = self.softmax_kernel_op(attn, block_bias, block_groups, batch_size, global_max,
                                                  output_scale=self.scale_output, output_dtype=torch.float8_e4m3fn)
        output = self.dequant_output(output_fp8)
        return output

    def forward_measure(self, attn, block_bias, block_groups, batch_size, global_max):
        measure_input((attn,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(attn, block_bias, block_groups, batch_size, global_max)
        measure_output((output,), self._mod_extra_config.outputs)
        return output


class PatchedLoRACompatibleLinear(PatchedLinearBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.init_linear(mod_extra_config)

    def forward_qdq(self, input, scale: float = 1.0):
        output = self.run_linear_qdq(input, self.bias)
        if self.lora_layer is not None:
            # TODO SW-174899 support lora layer quantization
            _raise_lora_layer_error(self.class_name_org)
        return output

    def forward_quant(self, input, scale: float = 1.0):
        output = self.run_linear_quant(input, self.bias)
        if self.lora_layer is not None:
            # TODO SW-174899 support lora layer quantization
            _raise_lora_layer_error(self.class_name_org)
        return output

    def forward_measure(self, input, scale: float = 1.0):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input, scale)
        measure_output((output,), self._mod_extra_config.outputs)
        return output


class PatchedLoRACompatibleConv(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        init_conv(self, mod_extra_config)

    def forward_qdq(self, input, scale: float = 1.0):
        qinput = self.quant_input(input)
        if self.lora_layer is not None:
            # TODO SW-174899 support lora layer quantization
            _raise_lora_layer_error(self.class_name_org)
        else:
            qweight = self.dequant_weights(self.weight, )
            output = torch.nn.functional.conv2d(
                qinput,
                qweight,
                self.bias,
                self.stride,
                self.padding,
                self.dilation,
                self.groups,
            )
        return output

    def forward_quant(self, input, scale: float = 1.0):
        qinput = self.quant_input(input)
        if self.lora_layer is not None:
            # TODO SW-174899 support lora layer quantization
            _raise_lora_layer_error(self.class_name_org)
        else:
            output = self.conv2d_fp8(qinput,
                                     self.weight,
                                     self.bias,
                                     self.stride,
                                     self.padding,
                                     self.dilation,
                                     self.groups,
                                     out_dtype=self._mod_extra_config.config_params["hp_dtype"],
                                     scale_input_inv=self.scale_input,
                                     scale_other_inv=self.scale_weight)
        return output

    def forward_measure(self, input, scale: float = 1.0):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input, scale)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight") if not self.use_qdq else "",
        )


class PatchedModuleFusedSDPA(PatchedModuleBase):
    def __init__(self, mod, parent, mod_extra_config, *args, **kwargs):
        # fsdpa is combined out of - BMM1(Q,K) -> Softmax -> BMM2(AMAX,V)
        # during measure we receive the amax value from the cguid and apply it during quant as input
        super().__init__(mod, parent, mod_extra_config, *args, **kwargs)
        self.fp8_fused_sdpa = get_quantized_func_wrapper(OP_TYPE.FSDPA, self.scale_format)
        if self.quantization_mode in [QuantMode.QUANTIZE, QuantMode.LOAD]:
            self.quant_q = self._mod_extra_config.inputs[0]
            self.quant_k = self._mod_extra_config.inputs[1]
            self.quant_v = self._mod_extra_config.inputs[2]
            self.dequant_output = self._mod_extra_config.outputs[0]
            self.register_scale("scale_q", mod_extra_config.scale.inputs[0].type(torch.float32), self.scale_format)
            self.register_scale("scale_k", mod_extra_config.scale.inputs[1].type(torch.float32), self.scale_format)
            self.register_scale("scale_v", mod_extra_config.scale.inputs[2].type(torch.float32), self.scale_format)
            self.register_scale("descale_amax", mod_extra_config.scale.inputs[3].type(
                torch.float32), self.scale_format)
            self.register_scale("scale_output", 1 /
                                mod_extra_config.scale.outputs[0].type(torch.float32), self.scale_format)
            self.register_scale("scale_amax", 1 / self.descale_amax, self.scale_format)

    def forward_qdq(
        self,
        q,
        k,
        v,
        attn_mask=None,
        dropout_p=0.0,
        is_causal=False,
        scale=None,
        softmax_mode="None",
        recompute=None,
        valid_seq_len=None,
        seq_padding_type="None",
    ):

        qinput = self.quant_q(q).detach()
        kinput = self.quant_k(k).detach()
        vinput = self.quant_v(v).detach()

        results = self._hpu_kernel_fsdpa.apply(
            qinput,
            kinput,
            vinput,
            attn_mask,
            dropout_p,
            is_causal,
            scale,
            softmax_mode,
            recompute,
            valid_seq_len,
            seq_padding_type,
        )
        return results

    def forward_quant(
        self,
        q,
        k,
        v,
        attn_mask=None,
        dropout_p=0.0,
        is_causal=False,
        scale=None,
        softmax_mode="None",
        recompute=None,
        valid_seq_len=None,
        seq_padding_type="None",
    ):
        sm_mode = softmax_mode if softmax_mode == "fp32" else "None"
        qinput = self.quant_q(q).detach()
        kinput = self.quant_k(k).detach()
        vinput = self.quant_v(v).detach()
        results = self.fp8_fused_sdpa(
            qinput,
            kinput,
            vinput,
            attn_mask=attn_mask,
            dropout_p=dropout_p,
            is_causal=is_causal,
            scale=scale,
            softmax_mode=sm_mode,
            d_scale_q=self.scale_q,
            d_scale_k=self.scale_k,
            d_scale_v=self.scale_v,
            q_scale_s=self.scale_amax,
            q_scale_o=self.scale_output,
            d_scale_s=self.descale_amax,
            is_amax_s=False,
            valid_seq_len=valid_seq_len,
            seq_padding_type=seq_padding_type,
        )
        output = results[0]
        d_out = self.dequant_output(output)
        return d_out

    def forward_measure(
        self,
        q,
        k,
        v,
        attn_mask=None,
        dropout_p=0.0,
        is_causal=False,
        scale=None,
        softmax_mode="fast",
        recompute=None,
        valid_seq_len=None,
        seq_padding_type="None",
    ):
        dq = q.detach()
        dk = k.detach()
        dv = v.detach()
        measure_input((dq, dk, dv), observer=self._mod_extra_config.inputs)
        results = self.fp8_fused_sdpa(
            dq,
            dk,
            dv,
            attn_mask=attn_mask,
            dropout_p=dropout_p,
            is_causal=is_causal,
            scale=scale,
            # fp8_fused_sdpa in bf16 can use either FastSoftmax or regular
            softmax_mode="fast",
            is_amax_s=True,
            valid_seq_len=valid_seq_len,
            seq_padding_type=seq_padding_type,
        )
        output = results[0]
        amax = results[1]
        measure_output((output, amax), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(
                self,
                "scale_q",
                "scale_k",
                "scale_v",
                "descale_amax",
                "scale_amax",
                "scale_output",
            ),
        )


class PatchedUnmeasuredModule(nn.Module):
    def __init__(self, name, mod, *args, **kwargs):
        super().__init__()
        if (name == "lm_head" and mod.__class__.__name__ == "ParallelLMHead"):
            # For LM_HEAD when using vLLM (as it doesn't use its forward, but a member's [quant_method] apply).
            self.__dict__.update(mod.__dict__)
            self.quant_method.apply = self.forward
        self.name = name

    def forward(self, *args, **kwargs):
        raise Exception(
            "Error - Layer '{}' was called but was not quantized because no measures were supplied.".format(self.name)
        )

    def extra_repr(self) -> str:
        return f"Dummy patch of {self.name} to raise exception as there are no measurements provided."



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import copy
import fnmatch
import json
import logging
import os
import random
import sys
import time
from dataclasses import dataclass, field
from typing import Dict, Optional, Sequence

import nltk
import numpy as np
import torch
import torch.nn.functional as F
import transformers
from datasets import load_dataset
from tqdm import tqdm
from transformers import AutoTokenizer

nltk.download("punkt", quiet=False)
random.seed(9973)

IGNORE_INDEX = -100
DEFAULT_PAD_TOKEN = "[PAD]"
DEFAULT_EOS_TOKEN = "</s>"
DEFAULT_BOS_TOKEN = "</s>"
DEFAULT_UNK_TOKEN = "</s>"
PROMPT_DICT = {
    "prompt_input": (
        "Below is an instruction that describes a task, paired with an input that provides further context. "
        "Write a response that appropriately completes the request.\n\n"
        "### Instruction:\n{instruction}\n\n Input:\n {article}\n\n### Response:"
    ),
    "prompt_no_input": (
        "Below is an instruction that describes a task. "
        "Write a response that appropriately completes the request.\n\n"
        "### Instruction:\n{instruction}\n\n### Response:"
    ),
}

current_path = os.path.abspath(__file__)
json_path = os.path.join(
    os.path.abspath(os.path.dirname(current_path) + os.path.sep + "."),
    "cnn_validation.json",
)


class DataCollatorForSupervisedDataset(object):
    """Collate examples for supervised fine-tuning."""

    def __init__(self, tokenizer: transformers.PreTrainedTokenizer):
        self.tokenizer = tokenizer

    def __call__(self, instances: Sequence[Dict]) -> Dict[str, torch.Tensor]:
        input_ids, labels = tuple([instance[key] for instance in instances] for key in ("input_ids", "labels"))
        input_ids = torch.nn.utils.rnn.pad_sequence(
            input_ids, batch_first=True, padding_value=self.tokenizer.pad_token_id
        )
        labels = torch.nn.utils.rnn.pad_sequence(labels, batch_first=True, padding_value=IGNORE_INDEX)
        return (input_ids, labels)


def _tokenize_fn(strings: Sequence[str], tokenizer: transformers.PreTrainedTokenizer) -> Dict:
    """Tokenize a list of strings."""
    tokenized_list = [
        tokenizer(
            text,
            return_tensors="pt",
            padding="max_length",
            max_length=tokenizer.model_max_length,
            truncation=True,
        )
        for text in strings
    ]
    input_ids = labels = [tokenized.input_ids[0] for tokenized in tokenized_list]
    input_ids_lens = labels_lens = [
        tokenized.input_ids.ne(tokenizer.pad_token_id).sum().item() for tokenized in tokenized_list
    ]
    return dict(
        input_ids=input_ids,
        labels=labels,
        input_ids_lens=input_ids_lens,
        labels_lens=labels_lens,
    )


def preprocess(
    sources: Sequence[str],
    targets: Sequence[str],
    tokenizer: transformers.PreTrainedTokenizer,
) -> Dict:
    """Preprocess the data by tokenizing."""
    examples = [s + t for s, t in zip(sources, targets)]
    examples_tokenized, sources_tokenized = [_tokenize_fn(strings, tokenizer) for strings in (examples, sources)]
    input_ids = examples_tokenized["input_ids"]
    labels = copy.deepcopy(input_ids)
    for label, source_len in zip(labels, sources_tokenized["input_ids_lens"]):
        label[:source_len] = IGNORE_INDEX

    return dict(input_ids=input_ids, labels=labels)


class CNNDAILYMAIL(object):
    def __init__(
        self,
        model_path,
        data_path=json_path,
        device="cpu",
        calib=False,
        num_samples=None,
    ):
        self.model_path = model_path
        self.data_path = data_path
        self.device = device
        self.num_samples = num_samples
        self.load_tokenizer()
        self.load_dataset()
        self.data_collator = DataCollatorForSupervisedDataset(tokenizer=self.tokenizer)
        self.calib = calib

    def wrap(self, example):
        example["instruction"] = "Summarize the following news article:"
        return example

    def load_dataset(self):
        """Loads dataset."""
        list_data_dict = load_dataset("cnn_dailymail", "3.0.0")["validation"]
        if self.num_samples is not None:
            self.num_samples = min(self.num_samples, len(list_data_dict))
            list_data_dict = random.choices(list_data_dict, k=self.num_samples)
        prompt_input, prompt_no_input = (
            PROMPT_DICT["prompt_input"],
            PROMPT_DICT["prompt_no_input"],
        )
        sources = [prompt_input.format_map(self.wrap(example)) for example in list_data_dict]
        targets = [f"{example['highlights']}" for example in list_data_dict]
        data_dict = preprocess(sources, targets, self.tokenizer)

        self.input_ids = data_dict["input_ids"]
        self.labels = data_dict["labels"]
        self.sources = sources
        self.targets = targets

    def load_tokenizer(self):
        """Returns the tokenizer."""
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_path,
            model_max_length=2048,
            padding_side="left",
            use_fast=False,
        )
        self.tokenizer.pad_token = self.tokenizer.eos_token

    def __len__(self):
        return len(self.input_ids)

    def __getitem__(self, i) -> Dict[str, torch.Tensor]:
        return dict(input_ids=self.input_ids[i], labels=self.labels[i])


def postprocess_text(preds, targets):
    preds = [pred.strip() for pred in preds]
    targets = [target.strip() for target in targets]

    # rougeLSum expects newline after each sentence
    # rougeLSum expects newline after each sentence
    preds = ["\n".join(nltk.sent_tokenize(pred)) for pred in preds]
    targets = ["\n".join(nltk.sent_tokenize(target)) for target in targets]

    return preds, targets



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# code ported from AutoGPTQ
import copy
import math
import os
import time
from typing import Dict, List, Optional, Union

import habana_frameworks.torch.core as htcore
import torch
import torch.nn as nn
import transformers

# from auto_gptq.quantization import Quantizer
from auto_gptq.modeling._base import BaseGPTQForCausalLM
from auto_gptq.modeling._const import CPU
from auto_gptq.modeling._utils import find_layers, get_device, get_module_by_name_prefix, move_to_device, pack_model
from auto_gptq.nn_modules._fused_base import FusedBaseAttentionModule, FusedBaseMLPModule
from auto_gptq.quantization import BaseQuantizeConfig
from auto_gptq.quantization.config import CHECKPOINT_FORMAT
from auto_gptq.utils.data_utils import collate_data
from transformers import PreTrainedModel

from neural_compressor.torch.algorithms.fp8_quant.utils.logger import logger

HPU = torch.device("hpu")


def quantize(x, scale, zero, maxq):
    if maxq < 0:
        return (x > scale / 2).float() * scale + (x < zero / 2).float() * zero
    q = torch.clamp(torch.round(x / scale) + zero, 0, maxq)
    return scale * (q - zero)


class HabanaQuantizer(nn.Module):
    def __init__(self, shape=1):
        super(HabanaQuantizer, self).__init__()
        self.register_buffer("maxq", torch.tensor(0))
        self.register_buffer("scale", torch.zeros(shape))
        self.register_buffer("zero", torch.zeros(shape))

    def configure(
        self,
        bits,
        perchannel=False,
        sym=True,
        mse=False,
        norm=2.4,
        grid=100,
        maxshrink=0.8,
        trits=False,
    ):
        self.maxq = torch.tensor(2**bits - 1)
        self.perchannel = perchannel
        self.sym = sym
        self.mse = mse
        self.norm = norm
        self.grid = grid
        self.maxshrink = maxshrink
        if trits:
            self.maxq = torch.tensor(-1)

    def find_params(self, x, weight=False):

        dev = x.device
        self.maxq = self.maxq.to(dev)

        shape = x.shape

        if self.perchannel:
            if weight:
                x = x.flatten(1)
            else:
                if len(shape) == 4:
                    x = x.permute([1, 0, 2, 3])
                    x = x.flatten(1)
                if len(shape) == 3:
                    x = x.reshape((-1, shape[-1])).t()
                if len(shape) == 2:
                    x = x.t()
        else:
            x = x.flatten().unsqueeze(0)

        tmp = torch.zeros(x.shape[0], device=dev)
        xmin = torch.minimum(x.min(1)[0], tmp)
        xmax = torch.maximum(x.max(1)[0], tmp)

        if self.sym:
            xmax = torch.maximum(torch.abs(xmin), xmax)
            tmp = xmin < 0
            if torch.any(tmp):
                # xmin[tmp] = -xmax[tmp]
                xmin[tmp] = (-xmax.cpu()[tmp.cpu()]).to(dev)
        tmp = (xmin == 0) & (xmax == 0)
        xmin[tmp] = -1
        xmax[tmp] = +1

        # htcore.mark_step()
        # print(f"a")

        if self.maxq < 0:
            self.scale = xmax
            self.zero = xmin
        else:
            self.scale = (xmax - xmin) / self.maxq
            if self.sym:
                self.zero = torch.full_like(self.scale, (self.maxq + 1) / 2)
            else:
                self.zero = torch.round(-xmin / self.scale)

        # htcore.mark_step()
        # print(f"b")

        if self.mse:
            best = torch.full([x.shape[0]], float("inf"), device=dev)
            for i in range(int(self.maxshrink * self.grid)):
                htcore.mark_step()
                print(f"Quantizer {i}")
                p = 1 - i / self.grid
                xmin1 = p * xmin
                xmax1 = p * xmax
                scale1 = (xmax1 - xmin1) / self.maxq
                zero1 = torch.round(-xmin1 / scale1) if not self.sym else self.zero
                q = quantize(x, scale1.unsqueeze(1), zero1.unsqueeze(1), self.maxq)
                q -= x
                q.abs_()
                q.pow_(self.norm)
                err = torch.sum(q, 1)
                tmp = err < best
                if torch.any(tmp):
                    best[tmp] = err[tmp]
                    self.scale[tmp] = scale1[tmp]
                    self.zero[tmp] = zero1[tmp]
        if not self.perchannel:
            if weight:
                tmp = shape[0]
            else:
                tmp = shape[1] if len(shape) != 3 else shape[2]
            self.scale = self.scale.repeat(tmp)
            self.zero = self.zero.repeat(tmp)

        if weight:
            shape = [-1] + [1] * (len(shape) - 1)
            self.scale = self.scale.reshape(shape)
            self.zero = self.zero.reshape(shape)
            return
        if len(shape) == 4:
            self.scale = self.scale.reshape((1, -1, 1, 1))
            self.zero = self.zero.reshape((1, -1, 1, 1))
        if len(shape) == 3:
            self.scale = self.scale.reshape((1, 1, -1))
            self.zero = self.zero.reshape((1, 1, -1))
        if len(shape) == 2:
            self.scale = self.scale.unsqueeze(0)
            self.zero = self.zero.unsqueeze(0)

    def quantize(self, x):
        if self.ready():
            return quantize(x, self.scale, self.zero, self.maxq)
        return x

    def enabled(self):
        return self.maxq > 0

    def ready(self):
        return torch.all(self.scale != 0)


class GaudiGPTQ:
    def __init__(self, layer):
        self.layer = layer
        self.dev = self.layer.weight.device
        W = layer.weight.data.clone()
        if isinstance(self.layer, nn.Conv2d):
            W = W.flatten(1)
        if isinstance(self.layer, transformers.pytorch_utils.Conv1D):
            W = W.t()
        self.rows = W.shape[0]
        self.columns = W.shape[1]
        self.H = torch.zeros((self.columns, self.columns), device=self.dev)
        self.nsamples = 0
        self.quantizer = HabanaQuantizer()

    def add_batch(self, inp, out):
        if os.environ.get("DEBUG"):
            self.inp1 = inp
            self.out1 = out
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if isinstance(self.layer, nn.Linear) or isinstance(self.layer, transformers.Conv1D):
            if len(inp.shape) == 3:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()
        if isinstance(self.layer, nn.Conv2d):
            unfold = nn.Unfold(
                self.layer.kernel_size,
                dilation=self.layer.dilation,
                padding=self.layer.padding,
                stride=self.layer.stride,
            )
            inp = unfold(inp)
            inp = inp.permute([1, 0, 2])
            inp = inp.flatten(1)
        self.H *= self.nsamples / (self.nsamples + tmp)
        self.nsamples += tmp
        # inp = inp.float()
        inp = math.sqrt(2 / self.nsamples) * inp.float()
        # self.H += 2 / self.nsamples * inp.matmul(inp.t())
        self.H += inp.matmul(inp.t())

    def fasterquant(
        self,
        blocksize=128,
        percdamp=0.01,
        group_size=-1,
        actorder=False,
        static_groups=False,
    ):
        W = self.layer.weight.data.clone()
        if isinstance(self.layer, nn.Conv2d):
            W = W.flatten(1)
        if isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        W = W.float()

        tick = time.time()

        htcore.mark_step()

        if not self.quantizer.ready():
            self.quantizer.find_params(W, weight=True)

        H = self.H.cpu()  # Keep Hessian in CPU
        del self.H
        dead = torch.diag(H) == 0
        H[dead, dead] = 1
        W[:, dead] = 0

        g_idx = []
        scale = []
        zero = []
        now_idx = 1

        if static_groups:
            import copy

            groups = []
            for i in range(0, self.columns, group_size):

                quantizer = copy.deepcopy(self.quantizer)
                quantizer.find_params(W[:, i : (i + group_size)], weight=True)
                scale.append(quantizer.scale)
                zero.append(quantizer.zero)
                groups.append(quantizer)

        if actorder:
            perm = torch.argsort(torch.diag(H), descending=True)
            W = W[:, perm]
            H = H[perm][:, perm]
            invperm = torch.argsort(perm)

        Losses = torch.zeros_like(W)
        Q = torch.zeros_like(W)

        damp = percdamp * torch.mean(torch.diag(H))
        diag = torch.arange(self.columns, device="cpu")

        H[diag, diag] += damp
        H = torch.linalg.cholesky(H)
        H = torch.cholesky_inverse(H)
        H = torch.linalg.cholesky(H, upper=True)
        Hinv = H

        for i1 in range(0, self.columns, blocksize):

            i2 = min(i1 + blocksize, self.columns)
            count = i2 - i1

            W1 = W[:, i1:i2].clone()
            Q1 = torch.zeros_like(W1)
            Err1 = torch.zeros_like(W1)
            Losses1 = torch.zeros_like(W1)
            Hinv1 = Hinv[i1:i2, i1:i2]

            for i in range(count):
                w = W1[:, i]
                d = Hinv1[i, i].to(HPU)

                if group_size != -1:
                    if not static_groups:
                        if (i1 + i) % group_size == 0:
                            self.quantizer.find_params(W[:, (i1 + i) : (i1 + i + group_size)], weight=True)

                        if ((i1 + i) // group_size) - now_idx == -1:
                            scale.append(self.quantizer.scale)
                            zero.append(self.quantizer.zero)
                            now_idx += 1
                    else:
                        idx = i1 + i
                        if actorder:
                            idx = perm[idx]
                        self.quantizer = groups[idx // group_size]

                q = self.quantizer.quantize(w.unsqueeze(1)).flatten()
                Q1[:, i] = q
                Losses1[:, i] = (w - q) ** 2 / d**2

                err1 = (w - q) / d
                W1[:, i:] -= err1.unsqueeze(1).matmul(Hinv1[i, i:].unsqueeze(0))
                Err1[:, i] = err1

            Q[:, i1:i2] = Q1
            Losses[:, i1:i2] = Losses1 / 2

            W[:, i2:] -= Err1.matmul(Hinv[i1:i2, i2:])

            if os.environ.get("DEBUG"):
                self.layer.weight.data[:, :i2] = Q[:, :i2]
                self.layer.weight.data[:, i2:] = W[:, i2:]
                logger.debug(torch.sum((self.layer(self.inp1) - self.out1) ** 2))
                logger.debug(torch.sum(Losses))

        torch.cuda.synchronize()
        logger.info(f"duration: {(time.time() - tick)}")
        logger.info(f"avg loss: {torch.sum(Losses).item() / self.nsamples}")

        group_size = group_size if group_size != -1 else self.columns
        if static_groups and actorder:
            g_idx = [perm[i] // group_size for i in range(self.columns)]
        else:
            g_idx = [i // group_size for i in range(self.columns)]
        g_idx = torch.tensor(g_idx, dtype=torch.int32, device=Q.device)
        if actorder:
            Q = Q[:, invperm]
            g_idx = g_idx[invperm]

        if isinstance(self.layer, transformers.Conv1D):
            Q = Q.t()
        self.layer.weight.data = Q.reshape(self.layer.weight.shape).type_as(self.layer.weight.data)
        if os.environ.get("DEBUG"):
            logger.debug(torch.sum((self.layer(self.inp1) - self.out1) ** 2))

        if scale == []:
            scale.append(self.quantizer.scale)
            zero.append(self.quantizer.zero)
        scale = torch.cat(scale, dim=1)
        zero = torch.cat(zero, dim=1)
        return scale, zero, g_idx

    def free(self):
        if os.environ.get("DEBUG"):
            self.inp1 = None
            self.out1 = None
        self.H = None
        self.Losses = None
        self.Trace = None
        torch.cuda.empty_cache()


def nested_move_to_device(v, device):
    if isinstance(v, torch.Tensor):
        return move_to_device(v, device)
    elif isinstance(v, (list, tuple)):
        return type(v)([nested_move_to_device(e, device) for e in v])
    else:
        return v


class BaseGaudiGPTQForCausalLM(BaseGPTQForCausalLM):
    layer_type: str = None
    layers_block_name: str = None
    outside_layer_modules: List[str] = None
    inside_layer_modules: List[List[str]] = None
    lm_head_name: str = "lm_head"

    fused_attn_module_type: Optional[FusedBaseAttentionModule] = None
    fused_mlp_module_type: Optional[FusedBaseMLPModule] = None

    def __init__(
        self,
        model: PreTrainedModel,
        quantized: bool,
        quantize_config: BaseQuantizeConfig,
        is_triton_backend: bool = False,
        injected_fused_attention: bool = False,
        injected_fused_mlp: bool = False,
        trainable: bool = False,
    ):
        super().__init__()

        # TODO: Due to logical errors, the code below is temporarily commented out.
        # def _convert_tensor_to_list(tensor):
        #     if isinstance(tensor, torch.Tensor):
        #         if len(tensor.shape) == 1:
        #             tensor = tensor.unsqueeze(0)
        #         tensor = tensor.long()
        #         return tensor.cpu().numpy().tolist()
        #     return [tensor]

        # new_examples = []
        # for example in examples:
        #     input_ids = _convert_tensor_to_list(example["input_ids"])
        #     attention_mask = _convert_tensor_to_list(example["attention_mask"])
        #     if "labels" in example:
        #         labels = _convert_tensor_to_list(example["labels"])
        #     elif "label" in example:
        #         labels = _convert_tensor_to_list(example["label"])
        #     elif "label_ids" in example:
        #         labels = _convert_tensor_to_list(example["label_ids"])
        #     else:
        #         labels = copy.deepcopy(input_ids)
        #     new_examples.append(
        #         {
        #             "input_ids": input_ids,
        #             "attention_mask": attention_mask,
        #             "labels": labels,
        #         }
        #     )
        # pad_token_id = self.config.pad_token_id
        # if not pad_token_id:
        #     pad_token_id = self.config.eos_token_id

        # new_examples = [
        #     collate_data(new_examples[start : start + batch_size], pad_token_id)
        #     for start in range(0, len(new_examples), batch_size)
        # ]
        # for new_example in new_examples:
        #     del new_example["labels"]

        # return new_examples

    @torch.inference_mode()
    def quantize(
        self,
        examples: List[Dict[str, Union[List[int], torch.LongTensor]]],
        batch_size: int = 1,
        use_triton: bool = False,
        use_cuda_fp16: bool = True,
        autotune_warmup_after_quantized: bool = False,
        cache_examples_on_gpu: bool = True,
    ):

        layer_inputs = []
        attention_masks = []
        position_ids = []
        layer_input_kwargs = []
        layer_outputs = []

        examples = self._prepare_examples_for_quantization(examples, batch_size)

        forward_pass_use_cache = self.model.config.use_cache
        self.model.config.use_cache = False

        num_batches = len(examples)
        layers = get_module_by_name_prefix(self.model, self.layers_block_name)

        cur_layer_device = get_device(layers[0])
        data_device = cur_layer_device if cache_examples_on_gpu else CPU

        def store_input_hook(_, args, kwargs):
            # Positional arguments.
            layer_input = []
            for inp in args:
                layer_input.append(move_to_device(inp, data_device))
            layer_inputs.append(layer_input)

            # Keyword arguments.
            if kwargs["attention_mask"] is not None:
                attention_masks.append(kwargs["attention_mask"].to(data_device))
            else:
                attention_masks.append(None)

            pos_ids = kwargs.get("position_ids", None)
            if pos_ids is not None:
                position_ids.append(move_to_device(pos_ids, data_device))
            one_kwargs = {}
            for (
                k,
                v,
            ) in kwargs.items():  # make sure other arguments also be captured
                if k not in ["hidden_states", "attention_mask", "position_ids"]:
                    one_kwargs[k] = nested_move_to_device(v, data_device)
            layer_input_kwargs.append(one_kwargs)
            raise ValueError

        force_layer_back_to_cpu = False
        if get_device(layers[0]) == CPU:
            layers[0] = layers[0].to(HPU)
            force_layer_back_to_cpu = True

        ori_outside_layer_module_devices = {}
        for module_name in self.outside_layer_modules:
            module = get_module_by_name_prefix(self.model, module_name)

            if module is None:
                continue

            ori_outside_layer_module_devices[module_name] = get_device(module)
            if module is not None:
                move_to_device(module, cur_layer_device)

        # TODO: make this optional, backporting https://github.com/huggingface/optimum/blob/main/optimum/gptq/quantizer.py
        handle = layers[0].register_forward_pre_hook(store_input_hook, with_kwargs=True)
        for example in examples:
            for k, v in example.items():
                if len(v.shape) == 1:
                    v = v.unsqueeze(0)
                example[k] = move_to_device(v, cur_layer_device)
            try:
                self.model(**example)
            except ValueError:
                pass
        handle.remove()

        move_to_device(layers[0], CPU if force_layer_back_to_cpu else cur_layer_device)
        for module_name in self.outside_layer_modules:
            module = get_module_by_name_prefix(self.model, module_name)
            if module is not None:
                move_to_device(module, ori_outside_layer_module_devices[module_name])

        torch.cuda.empty_cache()

        inside_layer_modules = self.inside_layer_modules
        if not self.quantize_config.true_sequential:
            inside_layer_modules = [sum(inside_layer_modules, [])]
        quantizers = {}
        for i in range(len(layers)):
            logger.info(f"Start quantizing layer {i + 1}/{len(layers)}")
            layer = layers[i]
            force_layer_back_to_cpu = False
            if get_device(layer) == CPU:
                move_to_device(layer, HPU)
                force_layer_back_to_cpu = True
            cur_layer_device = get_device(layer)

            full = find_layers(layer)
            for names in inside_layer_modules:
                subset = {n: full[n] for n in names if n in full}
                gptq = {}
                for name in subset:
                    gptq[name] = GaudiGPTQ(subset[name])
                    gptq[name].quantizer.configure(
                        self.quantize_config.bits,
                        perchannel=True,
                        sym=self.quantize_config.sym,
                        mse=False,
                    )

                def add_batch(name):
                    def tmp(_, inp, out):
                        # gptq is mutable.
                        gptq[name].add_batch(inp[0].data, out.data)  # noqa: F821

                    return tmp

                handles = []
                for name in subset:
                    handles.append(subset[name].register_forward_hook(add_batch(name)))
                for j in range(num_batches):
                    layer_input = []
                    for k, layer_inp in enumerate(layer_inputs[j]):
                        layer_input.append(move_to_device(layer_inp, cur_layer_device))

                    layer_attention_mask = move_to_device(attention_masks[j], cur_layer_device)
                    additional_layer_inputs = {"attention_mask": layer_attention_mask}
                    layer_position_ids = None if not position_ids else move_to_device(position_ids[j], cur_layer_device)
                    if layer_position_ids is not None:
                        additional_layer_inputs["position_ids"] = layer_position_ids
                    for k, v in layer_input_kwargs[j].items():
                        additional_layer_inputs[k] = nested_move_to_device(v, cur_layer_device)
                    layer(*layer_input, **additional_layer_inputs)
                for h in handles:
                    h.remove()

                for name in subset:
                    logger.info(f"Quantizing {name} in layer {i + 1}/{len(layers)}...")
                    scale, zero, g_idx = gptq[name].fasterquant(
                        percdamp=self.quantize_config.damp_percent,
                        group_size=self.quantize_config.group_size,
                        actorder=self.quantize_config.desc_act,
                        static_groups=self.quantize_config.static_groups,
                    )
                    quantizers[f"{self.layers_block_name}.{i}.{name}"] = (
                        gptq[name].quantizer.to(CPU if force_layer_back_to_cpu else cur_layer_device),
                        move_to_device(scale, CPU if force_layer_back_to_cpu else cur_layer_device),
                        move_to_device(zero, CPU if force_layer_back_to_cpu else cur_layer_device),
                        move_to_device(g_idx, CPU if force_layer_back_to_cpu else cur_layer_device),
                    )
                    gptq[name].free()

            for j in range(num_batches):
                layer_input = []
                for k, layer_inp in enumerate(layer_inputs[j]):
                    layer_input.append(move_to_device(layer_inp, cur_layer_device))

                layer_attention_mask = move_to_device(attention_masks[j], cur_layer_device)
                additional_layer_inputs = {"attention_mask": layer_attention_mask}
                layer_position_ids = None if not position_ids else move_to_device(position_ids[j], cur_layer_device)
                if layer_position_ids is not None:
                    additional_layer_inputs["position_ids"] = layer_position_ids
                for k, v in layer_input_kwargs[j].items():
                    additional_layer_inputs[k] = nested_move_to_device(v, cur_layer_device)
                layer_output = move_to_device(
                    layer(*layer_input, **additional_layer_inputs)[0],
                    cur_layer_device if cache_examples_on_gpu else CPU,
                )
                layer_outputs.append([layer_output])

            layers[i] = move_to_device(layer, CPU if force_layer_back_to_cpu else cur_layer_device)
            del layer
            del gptq
            del layer_inputs
            layer_inputs, layer_outputs = (
                layer_outputs,
                [],
            )  # TODO: is it really OK to cache only the first positional argument?
            torch.cuda.empty_cache()

        pack_model(
            model=self.model,
            quantizers=quantizers,
            bits=self.quantize_config.bits,
            group_size=self.quantize_config.group_size,
            use_triton=use_triton,
            use_cuda_fp16=use_cuda_fp16,
            desc_act=self.quantize_config.desc_act,
            warmup_triton=lambda enabled: None,
            force_layer_back_to_cpu=force_layer_back_to_cpu,
            use_marlin=self.quantize_config.checkpoint_format == CHECKPOINT_FORMAT.MARLIN,
        )
        # if device_map:
        #     self.model = remove_hook_from_module(self.model, recurse=True)
        #     self.model = simple_dispatch_model(self.model, device_map)
        self.model.config.use_cache = forward_pass_use_cache

        self._quantized = True

        torch.cuda.empty_cache()



# -*- coding: utf-8 -*-
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor PyTorch environment check."""

import importlib
import os
import sys
from functools import lru_cache

import torch
from packaging.version import Version

from neural_compressor.common.utils import logger

try:
    world_size = torch.distributed.get_world_size()
    local_rank = torch.distributed.get_rank()
except:
    local_rank = int(os.getenv("LOCAL_RANK", "-1"))
    world_size = int(os.getenv("WORLD_SIZE", "-1"))


################ Check imported sys.module first to decide behavior #################
def is_ipex_imported() -> bool:
    """Check whether intel_extension_for_pytorch is imported."""
    for name, _ in sys.modules.items():
        if name == "intel_extension_for_pytorch":
            return True
    return False


def is_transformers_imported() -> bool:
    """Check whether transformers is imported."""
    for name, _ in sys.modules.items():
        if name == "transformers":
            return True
    return False


################ Check available sys.module to decide behavior #################
def is_package_available(package_name):
    """Check if the package exists in the environment without importing.

    Args:
        package_name (str): package name
    """
    from importlib.util import find_spec

    package_spec = find_spec(package_name)
    return package_spec is not None


## check hpex
if is_package_available("habana_frameworks"):
    _hpex_available = True
    import habana_frameworks.torch.hpex  # pylint: disable=E0401
else:
    _hpex_available = False


def is_hpex_available():
    """Returns whether hpex is available."""
    return _hpex_available


def is_hpu_available():
    """Returns whether hpex is available."""
    return get_accelerator().name() == "hpu"


@lru_cache(None)
def is_hpex_support_g_idx():
    """Check if HPEX supports group_index in the schema of hpu::convert_from_int4."""
    if is_hpex_available():
        try:
            import habana_frameworks.torch
            import torch

            schema = torch._C._get_schema("hpu::convert_from_int4", "")
            return "group_index" in str(schema)
        except:  # pragma: no cover
            return False
    else:
        return False


## check optimum
if is_package_available("optimum"):
    _optimum_available = True
else:
    _optimum_available = False


def is_optimum_available():
    """Return whether optimum-habana is available."""
    return _optimum_available


## check optimum-habana
if is_package_available("optimum-habana"):
    if is_optimum_available():
        _optimum_habana_available = True
    else:
        raise ValueError("optimum-habana needs to be installed together with optimum.")
else:
    _optimum_habana_available = False


def is_optimum_habana_available():
    """Return whether optimum-habana is available."""
    return _optimum_habana_available


## check ipex
if is_package_available("intel_extension_for_pytorch"):
    _ipex_available = True
else:
    _ipex_available = False


def is_ipex_available():
    """Return whether ipex is available."""
    return _ipex_available


def get_ipex_version():
    """Return ipex version if ipex exists."""
    if is_ipex_available():
        try:
            import intel_extension_for_pytorch as ipex

            ipex_version = ipex.__version__.split("+")[0]
        except ValueError as e:  # pragma: no cover
            assert False, "Got an unknown version of intel_extension_for_pytorch: {}".format(e)
        version = Version(ipex_version)
        return version
    else:
        return None


TORCH_VERSION_2_2_2 = Version("2.2.2")
TORCH_VERSION_2_7_0 = Version("2.7.0")
TORCH_VERSION_2_8_0 = Version("2.8.0")


def get_torch_version():
    """Return torch version if ipex exists."""
    try:
        torch_version = torch.__version__.split("+")[0]
    except ValueError as e:  # pragma: no cover
        assert False, "Got an unknown version of torch: {}".format(e)
    version = Version(torch_version)
    return version


GT_OR_EQUAL_TORCH_VERSION_2_5 = get_torch_version() >= Version("2.5")


def get_accelerator(device_name="auto"):
    """Return the recommended accelerator based on device priority."""
    global accelerator  # update the global accelerator when calling this func
    from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

    accelerator = auto_detect_accelerator(device_name)
    inc_target_device = os.environ.get("INC_TARGET_DEVICE", None)
    if inc_target_device is not None and accelerator._name != inc_target_device.lower():
        auto_detect_accelerator.cache_clear()
        accelerator = auto_detect_accelerator(device_name)
    return accelerator


# for direct user access, used by @device_synchronize, can be changed by set_accelerator
accelerator = get_accelerator()


# for habana ease-of-use
def device_synchronize(raw_func):
    """Function decorator that calls accelerated.synchronize before and after a function call."""
    from functools import wraps

    @wraps(raw_func)
    def new_func(*args, **kwargs):
        accelerator.synchronize()
        output = raw_func(*args, **kwargs)
        accelerator.synchronize()
        return output

    return new_func


def can_pack_with_numba():  # pragma: no cover
    """Check if Numba and TBB are available for packing.

    To pack tensor with Numba, both Numba and TBB are required, and TBB should be configured correctly.
    """
    if not is_numba_available():
        logger.warning_once("Numba is not installed, please install it with `pip install numba`.")
        return False
    if not is_tbb_available():
        return False
    return True


def is_numba_available():  # pragma: no cover
    """Check if Numba is available."""
    try:
        import numba

        return True
    except ImportError:
        return False


def _is_tbb_installed():  # pragma: no cover
    import importlib.metadata

    try:
        importlib.metadata.version("tbb")
        return True
    except importlib.metadata.PackageNotFoundError:
        return False


def _is_tbb_configured():  # pragma: no cover
    try:
        from numba.np.ufunc.parallel import _check_tbb_version_compatible

        # check if TBB is present and compatible
        _check_tbb_version_compatible()

        return True
    except ImportError as e:
        logger.warning_once(f"TBB not available: {e}")
        return False


def is_tbb_available():  # pragma: no cover
    """Check if TBB is available."""
    if not _is_tbb_installed():
        logger.warning_once("TBB is not installed, please install it with `pip install tbb`.")
        return False
    if not _is_tbb_configured():
        logger.warning_once(
            (
                "TBB is installed but not configured correctly. \n"
                "Please add the TBB library path to `LD_LIBRARY_PATH`, "
                "for example: `export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/usr/local/lib/`."
            )
        )
        return False
    return True


def get_used_hpu_mem_MB():
    """Get HPU used memory: MiB."""
    import numpy as np
    from habana_frameworks.torch.hpu import memory_stats

    torch.hpu.synchronize()
    mem_stats = memory_stats()
    used_hpu_mem = np.round(mem_stats["InUse"] / 1024**2, 3)
    return used_hpu_mem


def get_used_cpu_mem_MB():
    """Get the amount of CPU memory used by the current process in MiB (Mebibytes)."""
    import psutil

    process = psutil.Process()
    mem_info = process.memory_info()
    used_cpu_mem = round(mem_info.rss / 1024**2, 3)
    return used_cpu_mem



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from ..fp_utils import get_fp8_hw_alligned_scales
from ...utils.logger import logger
from enum import Enum, auto
import re
import json

class ScaleMethodString(Enum):
    UNIT_SCALE = auto()
    HW_ALIGNED_SINGLE_SCALE = auto()
    MAXABS_HW = auto()
    MAXABS_POW2 = auto()
    ACT_MAXABS_HW_WEIGHTS_PCS_MAXABS_POW2 = auto()
    ACT_MAXABS_HW_WEIGHTS_PCS_OPT_POW2 = auto()
    ACT_MAXABS_POW2_WEIGHTS_PCS_MAXABS_POW2 = auto()
    ACT_MAXABS_POW2_WEIGHTS_PCS_OPT_POW2 = auto()
    MAXABS_HW_OPT_WEIGHT = auto()
    MAXABS_POW2_OPT_WEIGHT = auto()
    MAXABS_ARBITRARY = auto()
    ACT_MAXABS_PCS_POW2_WEIGHT_MAXABS_PTS_POW2_HW = auto()

class ScaleGranularity(Enum):
    PTS = auto()
    PCS = auto()

class ScaleValueType(Enum):
    MAXABS = auto()
    FIXED_VALUE = auto()
    OPT = auto()
    DUMMY_SCALES = auto()

class ScaleRoundMethod(Enum):
    IDENTITY = auto()
    POW2 = auto()
    HW_ALIGNED = auto()
    HW_ALIGNED_FIXED = auto()
    SCALE_UNIT = auto()

class CfgStr(Enum):
    ACTIVATION = "activation"
    WEIGHT = "weight"
    DEFAULT = "default"
    NODES= "nodes"
    LAYERS= "layers"
    LAYER_TYPES= "layer_types"
    GRANULARITY = "granularity"
    SCALE_VALUE_TYPE = "scale_value_type"
    ROUNDING_METHOD = "rounding_method"
    BACKOFF = "backoff"
    PARAMS = "params"
    LAYERS_DOT_PATTERN= r"layers\.(\d+)"
    LAYERS_SLASH_PATTERN= r"layers/(\d+)"

class ScaleMethodConfig:
    def __init__(self, 
                 granularity=ScaleGranularity.PTS,
                 scale_value_type=ScaleValueType.MAXABS,
                 rounding_method=ScaleRoundMethod.IDENTITY,
                 backoff=1.0,
                 params=None):
        self.granularity = granularity
        self.scale_value_type = scale_value_type
        self.rounding_method = rounding_method
        self.backoff = backoff
        self.params = params if params is not None else {}

    def __repr__(self):
        return f"ScaleMethodConfig(granularity={self.granularity}, scale_value_type={self.scale_value_type}, rounding_method={self.rounding_method}, backoff={self.backoff}, params={self.params})"

    def __hash__(self):
        # TODO: For supporting a user custom ScaleMethodConfig, we need to include the backoff and the params in the hash.
        return hash((
            self.granularity,
            self.scale_value_type,
            self.rounding_method
        ))
    
    def __eq__(self, other):
        if not isinstance(other, ScaleMethodConfig):
            return False
        
        # Only check the three fields that define uniqueness
        return (self.granularity == other.granularity and
                self.scale_value_type == other.scale_value_type and
                self.rounding_method == other.rounding_method)

scale_method_config_mapping = {
    ScaleMethodString.UNIT_SCALE:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(scale_value_type=ScaleValueType.FIXED_VALUE, rounding_method= ScaleRoundMethod.SCALE_UNIT),
        CfgStr.ACTIVATION: ScaleMethodConfig(scale_value_type=ScaleValueType.FIXED_VALUE, rounding_method= ScaleRoundMethod.SCALE_UNIT)
    },
    ScaleMethodString.HW_ALIGNED_SINGLE_SCALE:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(scale_value_type=ScaleValueType.FIXED_VALUE, rounding_method= ScaleRoundMethod.HW_ALIGNED_FIXED),
        CfgStr.ACTIVATION: ScaleMethodConfig(scale_value_type=ScaleValueType.FIXED_VALUE, rounding_method= ScaleRoundMethod.HW_ALIGNED_FIXED)
    },
    ScaleMethodString.MAXABS_HW:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(rounding_method= ScaleRoundMethod.HW_ALIGNED, backoff= 0.5),
        CfgStr.ACTIVATION: ScaleMethodConfig(rounding_method= ScaleRoundMethod.HW_ALIGNED, backoff= 0.25)
    },
    ScaleMethodString.MAXABS_POW2:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(rounding_method= ScaleRoundMethod.POW2, backoff= 0.5),
        CfgStr.ACTIVATION: ScaleMethodConfig(rounding_method= ScaleRoundMethod.POW2, backoff= 0.25)
    },
    ScaleMethodString.MAXABS_ARBITRARY:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(backoff= 0.5),
        CfgStr.ACTIVATION: ScaleMethodConfig(backoff= 0.25)
    },
    ScaleMethodString.ACT_MAXABS_PCS_POW2_WEIGHT_MAXABS_PTS_POW2_HW:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(rounding_method= ScaleRoundMethod.HW_ALIGNED, backoff= 0.5),
        CfgStr.ACTIVATION: ScaleMethodConfig(granularity= ScaleGranularity.PCS, rounding_method= ScaleRoundMethod.POW2)
    },
    ScaleMethodString.MAXABS_HW_OPT_WEIGHT:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(scale_value_type= ScaleValueType.OPT, rounding_method= ScaleRoundMethod.HW_ALIGNED, backoff= 0.5, params={"weight_scales": get_fp8_hw_alligned_scales(torch.float8_e4m3fn)}),
        CfgStr.ACTIVATION: ScaleMethodConfig(rounding_method= ScaleRoundMethod.HW_ALIGNED, backoff= 0.25)
    },
    ScaleMethodString.MAXABS_POW2_OPT_WEIGHT:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(scale_value_type= ScaleValueType.OPT, rounding_method= ScaleRoundMethod.POW2, backoff= 0.5, params={"weight_scales": [2.0**s for s in range(-10, 10)]}),
        CfgStr.ACTIVATION: ScaleMethodConfig(rounding_method= ScaleRoundMethod.POW2, backoff= 0.25)
    },
    ScaleMethodString.ACT_MAXABS_HW_WEIGHTS_PCS_MAXABS_POW2:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(granularity= ScaleGranularity.PCS, rounding_method= ScaleRoundMethod.POW2, backoff= 0.5),
        CfgStr.ACTIVATION: ScaleMethodConfig(rounding_method= ScaleRoundMethod.HW_ALIGNED, backoff= 0.25)
    },
    ScaleMethodString.ACT_MAXABS_HW_WEIGHTS_PCS_OPT_POW2:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(scale_value_type = ScaleValueType.OPT, granularity= ScaleGranularity.PCS, rounding_method= ScaleRoundMethod.POW2, backoff= 0.5, params={"weight_scales": [2.0**s for s in range(-3, 5)]}),
        CfgStr.ACTIVATION: ScaleMethodConfig(rounding_method= ScaleRoundMethod.HW_ALIGNED, backoff= 0.25)
    },
    ScaleMethodString.ACT_MAXABS_POW2_WEIGHTS_PCS_MAXABS_POW2:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(granularity= ScaleGranularity.PCS, rounding_method= ScaleRoundMethod.POW2, backoff= 0.5),
        CfgStr.ACTIVATION: ScaleMethodConfig(rounding_method= ScaleRoundMethod.POW2, backoff= 0.25)
    },
    ScaleMethodString.ACT_MAXABS_POW2_WEIGHTS_PCS_OPT_POW2:
    {
        CfgStr.WEIGHT:     ScaleMethodConfig(scale_value_type = ScaleValueType.OPT, granularity= ScaleGranularity.PCS, rounding_method= ScaleRoundMethod.POW2, backoff= 0.5, params={"weight_scales": [2.0**s for s in range(-3, 5)]}),
        CfgStr.ACTIVATION: ScaleMethodConfig(rounding_method= ScaleRoundMethod.POW2, backoff= 0.25)
    },
}

reverse_scale_method_mapping = {
    (configs[CfgStr.WEIGHT], configs[CfgStr.ACTIVATION]): scale_method
    for scale_method, configs in scale_method_config_mapping.items()
}

def get_scale_method_from_config(config_dict):
    """
    Given a dictionary with 'weight' and 'activation' ScaleMethodConfig objects,
    returns the corresponding ScaleMethodString if it matches a known configuration,
    otherwise returns None.
    """
    weight_config = config_dict[CfgStr.WEIGHT]
    activation_config = config_dict[CfgStr.ACTIVATION]
    config_key = (weight_config, activation_config)
    return reverse_scale_method_mapping.get(config_key, None)

def dict_to_scale_method_config(scale_method, scale_method_config_default=None):
    """
    Converts a dictionary specifying scale method parameters (granularity, scale_value_type,
    rounding_method, backoff, params) into a ScaleMethodConfig object.
    Tries to get each parameter from scale_method, then from scale_method_config_default, then uses the hardcoded default.
    Always converts string enum values to enum objects.
    """
    def get_param(key, default_value):
        if key in scale_method:
            return scale_method[key]
        elif scale_method_config_default is not None and hasattr(scale_method_config_default, key):
            return getattr(scale_method_config_default, key)
        else:
            return default_value
    scale_method_config = ScaleMethodConfig(
        granularity=get_param(CfgStr.GRANULARITY.value, ScaleGranularity.PTS),
        scale_value_type=get_param(CfgStr.SCALE_VALUE_TYPE.value, ScaleValueType.MAXABS),
        rounding_method=get_param(CfgStr.ROUNDING_METHOD.value, ScaleRoundMethod.IDENTITY),
        backoff=get_param(CfgStr.BACKOFF.value, 1.0),
        params=get_param(CfgStr.PARAMS.value, None),
    )
    return scale_method_config

def find_node_scale_method_config(scale_method_config, node_name, layer_type):
    """
    Finds and returns the appropriate scale method config for a given node name and layer type.
    Checks nodes, layers (by index), and layer_types in order, falling back to the default if not found.
    """
    # Try node-specific config
    scale_method_config_nodes = scale_method_config.get(CfgStr.NODES, None)
    if scale_method_config_nodes and scale_method_config_nodes.get(node_name, None) is not None:
        logger.trace(f"Scale method config node specific found for node: {node_name}")
        return scale_method_config_nodes[node_name]

    # Try layer index config (dot or slash pattern)
    match = re.search(CfgStr.LAYERS_DOT_PATTERN.value, node_name) # Match ".layers.<number>."
    if not match:
        match = re.search(CfgStr.LAYERS_SLASH_PATTERN.value, node_name) # Match "/layers/<number>/"
    if match:
        layer_index = match.group(1)
        scale_method_config_layers = scale_method_config.get(CfgStr.LAYERS, None)
        if scale_method_config_layers and scale_method_config_layers.get(layer_index, None) is not None:
            logger.trace(f"Scale method config found for layer index: {layer_index} for node: {node_name}")
            return scale_method_config_layers[layer_index]

    # Try layer type config
    scale_method_config_layer_types = scale_method_config.get(CfgStr.LAYER_TYPES, None)
    if scale_method_config_layer_types and scale_method_config_layer_types.get(layer_type, None) is not None:
        logger.trace(f"Scale method config found for layer type: {layer_type} for node: {node_name}")
        return scale_method_config_layer_types[layer_type]

    # Fallback to default config
    logger.trace(f"Using default scale method config for node: {node_name}, layer_type: {layer_type}")
    return scale_method_config[CfgStr.DEFAULT]

def _check_config(
    config,
    scale_value_type_weight=None,
    scale_value_type_activation=None,
    rounding_method_weight=None,
    rounding_method_activation=None,
    granularity_weight=None,
    granularity_activation=None,
):
    """
    Checks if the given config matches all specified criteria for weight and activation
    (scale_value_type, rounding_method, granularity). Returns True if all match or are None.
    """
    checks = [
        (scale_value_type_weight is None or config[CfgStr.WEIGHT].scale_value_type == scale_value_type_weight,
         f"scale_value_type_weight: expected {scale_value_type_weight}, got {config[CfgStr.WEIGHT].scale_value_type}"),
        (scale_value_type_activation is None or config[CfgStr.ACTIVATION].scale_value_type == scale_value_type_activation,
         f"scale_value_type_activation: expected {scale_value_type_activation}, got {config[CfgStr.ACTIVATION].scale_value_type}"),
        (rounding_method_weight is None or config[CfgStr.WEIGHT].rounding_method == rounding_method_weight,
         f"rounding_method_weight: expected {rounding_method_weight}, got {config[CfgStr.WEIGHT].rounding_method}"),
        (rounding_method_activation is None or config[CfgStr.ACTIVATION].rounding_method == rounding_method_activation,
         f"rounding_method_activation: expected {rounding_method_activation}, got {config[CfgStr.ACTIVATION].rounding_method}"),
        (granularity_weight is None or config[CfgStr.WEIGHT].granularity == granularity_weight,
         f"granularity_weight: expected {granularity_weight}, got {config[CfgStr.WEIGHT].granularity}"),
        (granularity_activation is None or config[CfgStr.ACTIVATION].granularity == granularity_activation,
         f"granularity_activation: expected {granularity_activation}, got {config[CfgStr.ACTIVATION].granularity}"),
    ]
    for passed, msg in checks:
        if not passed:
            logger.trace(f"Scale method config validation failed: {msg}")
            return False
    return True

def check_scale_method_fields(
    scale_method_dict,
    scale_method = None,
    scale_value_type_weight=None,
    scale_value_type_activation=None,
    rounding_method_weight=None,
    rounding_method_activation=None,
    granularity_weight=None,
    granularity_activation=None,
    reducer=all,
):
    """
    Iterates over all scale method configs (default, nodes, layers, layer_types) in the given dict and compares them to the given conditions.
    For each config, verifies that the specified fields match the provided values (if not None).
    If scale_method is given, uses its config as reference for the checks.
    The reducer argument determines if all (default) or any must match (use reducer=any).
    Returns the result of the reducer over all configs.
    """
    if scale_method is not None:
        scale_method_config = scale_method_config_mapping[scale_method]
        scale_method_config_weight = scale_method_config[CfgStr.WEIGHT]
        scale_method_config_activation = scale_method_config[CfgStr.ACTIVATION]
        scale_value_type_weight = scale_method_config_weight.scale_value_type
        scale_value_type_activation = scale_method_config_activation.scale_value_type
        rounding_method_weight = scale_method_config_weight.rounding_method
        rounding_method_activation = scale_method_config_activation.rounding_method
        granularity_weight = scale_method_config_weight.granularity
        granularity_activation = scale_method_config_activation.granularity

    configs_to_check = []
    for key in [CfgStr.DEFAULT, CfgStr.NODES, CfgStr.LAYERS, CfgStr.LAYER_TYPES]:
        if key in scale_method_dict:
            configs = scale_method_dict[key]
            if isinstance(configs, dict) and key is not CfgStr.DEFAULT:
                configs_to_check.extend(configs.values())
            else:
                configs_to_check.append(configs)
    return reducer(_check_config(config, scale_value_type_weight, scale_value_type_activation, rounding_method_weight, rounding_method_activation, granularity_weight, granularity_activation) for config in configs_to_check)

def dump_scale_method_config_by_mod_map(scale_method_config_by_mod_map, filename="scale_method_config_by_mod_map.json"):
    """
    Dumps the scale_method_config_by_mod_map to a JSON file.
    """
    serializable = {}
    for mod_name, cfgs in scale_method_config_by_mod_map.items():
        serializable[mod_name] = {}
        for key, cfg in cfgs.items():
            serializable[mod_name][ key.name if hasattr(key, "name") else str(key) ] = {
                    CfgStr.GRANULARITY.value: cfg.granularity.name,
                    CfgStr.SCALE_VALUE_TYPE.value: cfg.scale_value_type.name,
                    CfgStr.ROUNDING_METHOD.value: cfg.rounding_method.name,
                    CfgStr.BACKOFF.value: cfg.backoff,
                    CfgStr.PARAMS.value: cfg.params,
            }
    with open(filename, "w") as f:
        json.dump(serializable, f, indent=2)

def load_scale_method_config_by_mod_map(filename):
    """
    Loads scale_method_config_by_mod_map from a JSON file and reconstructs the original structure.
    Returns: dict[str, dict[CfgStr, ScaleMethodConfig]]
    """
    with open(filename, "r") as f:
        data = json.load(f)
    result = {}
    for mod_name, cfgs in data.items():
        result[mod_name] = {}
        for key, cfg_dict in cfgs.items():
            enum_key = CfgStr[key.upper()]
            result[mod_name][enum_key] = ScaleMethodConfig(
                granularity=ScaleGranularity[cfg_dict[CfgStr.GRANULARITY.value]],
                scale_value_type=ScaleValueType[cfg_dict[CfgStr.SCALE_VALUE_TYPE.value]],
                rounding_method=ScaleRoundMethod[cfg_dict[CfgStr.ROUNDING_METHOD.value]],
                backoff=cfg_dict.get(CfgStr.BACKOFF.value, 1.0),
                params=cfg_dict.get(CfgStr.PARAMS.value, None),
            )
    return result





# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import abstractmethod
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator
import torch
from .round_scales_function import ScaleIdentity
# TODO [SW-224612]: Use cguid to calc scales and remove is_calc_scale_with_cguid
from ..common import QuantTensorType, is_calc_scale_with_cguid
from ..fp_utils import mmse_scale_multi, get_fullscale, mmse_scale, calc_scale_from_maxabs, invert_scale, calculate_scale_maxabs_with_cguid, ScaleCalculationMaxMode
from ...utils.logger import logger

class ScalesMethod:

    curr_device = auto_detect_accelerator()

    def __init__(self, round_scale_method, params, device_for_scale, fullscale=None, is_dynamic=False):
        self.round_scale_method = round_scale_method
        self.params = params
        self.hp_dtype = self.params["hp_dtype"]
        self.lp_dtype = self.params["lp_dtype"]
        self.device = torch.device(self.curr_device.name())
        self.fullscale = fullscale if fullscale is not None else get_fullscale(self.lp_dtype, device_for_scale)
        self.scale = None
        self.is_dynamic = is_dynamic
        #TODO [SW-224612]: Use cguid to calc scales and remove check
        #TODO [SW-239725]: Re-enable scale calculation in cguid for static quantization
        self.calc_scale_with_cguid = is_calc_scale_with_cguid() if is_dynamic else False
        logger.trace("%s %s", self.__class__.__name__, self.__dict__)

    def __repr__(self):
        return f"{self.__class__.__name__}(round_scale_method={self.round_scale_method.__class__.__name__}, params={self.params}, " \
            f"fullscale={self.fullscale}, scale={self.scale}, is_dynamic={self.is_dynamic})"

    def calc_invert_scales(self):
        return self.invert_scales(self.scale)

    @staticmethod
    def invert_scales(scale=None):
        return invert_scale(scale)

    @abstractmethod
    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        raise NotImplementedError("`calc_scales` function is not implemented")


class MaxAbsMethod(ScalesMethod):
    def __init__(self, round_scale_method, params, device_for_scales, backoff, fullscale=None, is_dynamic=False):
        super().__init__(round_scale_method, params, device_for_scales, fullscale, is_dynamic)
        self.backoff = backoff
        self.calc_scale_func_dict = self.get_scale_funcs_dict()
        logger.trace("%s %s", self.__class__.__name__, self.__dict__)

    def get_scale_funcs_dict(self):
        return {QuantTensorType.CONST: self.calc_scale_from_const_tensor,
                QuantTensorType.MEASUREMENTS: self.calc_scale_from_measurement}

    @abstractmethod
    def calc_scale_from_const_tensor(self, tensor):
        raise NotImplementedError("`calc_scale_from_const_tensor` function is not implemented")

    def calc_scale_from_measurement(self, tensor):
        # TODO [SW-235427]: Check if we need to remove max()
        maxabs_tensor =  torch.tensor(tensor, dtype=self.hp_dtype, device=self.device).max()
        scale_tensor = calc_scale_from_maxabs(maxabs_tensor, self.fullscale, self.backoff)
        return scale_tensor

    def _calculate_maxabs_scale(self, tensor, tensor_type):
        scale_tensor = self.calc_scale_func_dict[tensor_type](tensor)
        scale = self.round_scale_method.calc(scale_tensor)
        return scale

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        self.scale = self._calculate_maxabs_scale(tensor, tensor_type)
        return self.scale

class MaxAbsPts(MaxAbsMethod):
    def __init__(self, round_scale_method, params, device_for_scales, backoff, fullscale=None, is_dynamic=False):
        super().__init__(round_scale_method, params, device_for_scales, backoff, fullscale, is_dynamic)
        logger.trace("%s %s", self.__class__.__name__, self.__dict__)

    def calc_scale_from_const_tensor(self, tensor):
        # TODO [SW-224612]: Use cguid to calc scales and remove check
        if self.calc_scale_with_cguid:
            # TODO: [SW-233670] check why self.hp_dtype conversion is necessary here, and consider moving it to cguid
            scale_tensor = calculate_scale_maxabs_with_cguid(
                tensor, ScaleCalculationMaxMode.MAX_ABS_PTS_CALCULATION, fullscale=self.fullscale, backoff=self.backoff
            ).to(self.hp_dtype)
        else:
            maxabs_tensor = torch.max(torch.abs(tensor.detach())).to(dtype=self.hp_dtype, device=self.device)
            scale_tensor = calc_scale_from_maxabs(maxabs_tensor, self.fullscale, self.backoff)
        return scale_tensor


## MulAdditionalScales Get 2 input scales, and return their multiplication.
# used for linear and matmul outputs
class MulAdditionalScales(ScalesMethod):
    def __init__(self, round_scale_method, params, device_for_scales, is_dynamic=False):
        super().__init__(round_scale_method, params, device_for_scales, is_dynamic=is_dynamic)

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        input0 = additional_kwargs["input0"]
        input1 = additional_kwargs["input1"]
        self.scale = input0 * input1
        return self.scale


class MulAdditionalDynamicScales(MulAdditionalScales):
    def __init__(self, round_scale_method, params, device_for_scales):
        super().__init__(round_scale_method, params, device_for_scales, is_dynamic=True)

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        if tensor_type != QuantTensorType.DYNAMIC:
            return None
        return super().calc_scales(tensor, tensor_type, **additional_kwargs)


## UseFirstAdditionalScales Get 2 input scales, and return the first one.
# used for linear smooth quant output
class UseFirstAdditionalScales(ScalesMethod):
    def __init__(self, round_scale_method, params, device_for_scales):
        super().__init__(round_scale_method, params, device_for_scales)

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        input0 = additional_kwargs["input0"]
        self.scale = input0
        return self.scale

## DummyScales always return 1.0 as scale,
# used when running with dummy measurement (prepare_model_with_dummy_measurement)
class DummyScales(ScalesMethod):
    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        self.scale = torch.tensor(1.0).to(self.device)
        return self.scale


class MaxAbsPcs(MaxAbsMethod):

    def __init__(self, round_scale_method, params, device_for_scales, backoff, fullscale=None, dim=1, keepdim=False, is_dynamic=False):
        super().__init__(round_scale_method, params, device_for_scales, backoff, fullscale, is_dynamic=is_dynamic)
        self.dim = dim
        self.keepdim = keepdim
        self.eps = torch.tensor(torch.finfo(torch.bfloat16).tiny, device=self.device)
        logger.trace("%s %s", self.__class__.__name__, self.__dict__)

    def calc_scale_from_const_tensor_no_reshape(self, tensor):
        #TODO [SW-224612]: Use cguid to calc scales and remove check
        if self.calc_scale_with_cguid:
            scale_tensor = calculate_scale_maxabs_with_cguid(
                tensor,
                ScaleCalculationMaxMode.MAX_ABS_PCS_CALCULATION,
                reduceAxis=self.dim,
                reduceKeepdim=self.keepdim,
                fullscale=self.fullscale,
                backoff=self.backoff,
            )
        else:
            maxabs_tensor = torch.amax(torch.abs(tensor), dim=self.dim, keepdim=self.keepdim)
            scale_tensor = calc_scale_from_maxabs(maxabs_tensor, self.fullscale, self.backoff)
            scale_tensor = torch.max(scale_tensor, self.eps)
        return scale_tensor

    def calc_scale_from_const_tensor(self, tensor):
        scale_tensor = self.calc_scale_from_const_tensor_no_reshape(tensor).reshape([-1, 1])
        if self.dim == 1:
            scale_tensor = scale_tensor.flatten()
        return scale_tensor


## InputChannelScale used for input channel in PCS mode
class InputChannelScale(ScalesMethod):
    def __init__(self, round_scale_method, params, device_for_scales, in_channel_size):
        super().__init__(round_scale_method, params, device_for_scales)
        self.in_channel_size = in_channel_size

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        input_in_ch = torch.ones([self.in_channel_size, 1], dtype=self.hp_dtype, device=self.device)
        return  input_in_ch.flatten()

class FixedScale(ScalesMethod):
    
    def __init__(self, round_scale_method, params, device_for_scales):
        super().__init__(round_scale_method, params, device_for_scales)
        self.round_scale_method = round_scale_method

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        self.scale = torch.tensor(self.round_scale_method.calc(tensor), dtype=self.hp_dtype, device=self.device)
        return self.scale


class OptScalesPts(ScalesMethod):

    def __init__(self, round_scale_method, optional_scales_list, params, device_for_scales, backoff):
        super().__init__(round_scale_method, params, device_for_scales)
        self.round_scale_method = round_scale_method
        self.optional_scales_list = optional_scales_list
        self.backoff = backoff

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        self.scale = self.round_scale_method.calc(mmse_scale(tensor, self.optional_scales_list, self.lp_dtype, self.hp_dtype))
        return self.scale

class OptScalesPcs(ScalesMethod):
    def __init__(self, round_scale_method, optional_scales_list, params, device_for_scales, backoff):
        super().__init__(round_scale_method, params, device_for_scales)
        self.round_scale_method = round_scale_method
        self.optional_scales_list = optional_scales_list
        self.backoff = backoff
        self.device_for_scales = device_for_scales

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        maxabs_scales_as_ref = MaxAbsPcs(self.round_scale_method, self.params, self.device_for_scales,
                                         self.backoff).calc_scales(tensor, tensor_type)
        const_opt_scale_out_ch = mmse_scale_multi(
            torch.transpose(tensor, 0, 1),
            maxabs_scales_as_ref,
            self.optional_scales_list,
            self.lp_dtype,
            self.hp_dtype,
        ).unsqueeze(1)
        self.scale = self.round_scale_method.calc(const_opt_scale_out_ch).flatten()
        return self.scale


class MaxAbsDynamicPcs(MaxAbsPcs):

    def __init__(self, round_scale_method, params, device_for_scales, backoff, fullscale=None):
        super().__init__(round_scale_method, params, device_for_scales, backoff, fullscale, -1, True, True)
        logger.trace("%s %s", self.__class__.__name__, self.__dict__)

    def get_scale_funcs_dict(self):
        scale_funcs_dict = super().get_scale_funcs_dict()
        scale_funcs_dict[QuantTensorType.DYNAMIC] = self.calc_scale_from_const_tensor_no_reshape
        return scale_funcs_dict

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        # In dynamic quantization the scale is changed each time,
        # and setting scale as a member is not supported in hpu graphs and torch.compile
        # (it can break the graph)
        return self._calculate_maxabs_scale(tensor, tensor_type)


class MaxAbsDynamicPts(MaxAbsPts):

    def __init__(self, round_scale_method, params, device_for_scales, backoff, fullscale=None):
        super().__init__(round_scale_method, params, device_for_scales, backoff, fullscale, True)
        logger.trace("%s %s",self.__class__.__name__, self.__dict__)

    def get_scale_funcs_dict(self):
        scale_funcs_dict = super().get_scale_funcs_dict()
        scale_funcs_dict[QuantTensorType.DYNAMIC] = self.calc_scale_from_const_tensor
        return scale_funcs_dict

    def calc_scales(self, tensor, tensor_type, **additional_kwargs):
        # In dynamic quantization the scale is changed each time,
        # and setting scale as a member is not supported in hpu graphs and torch.compile
        # (it can break the graph)
        return self._calculate_maxabs_scale(tensor, tensor_type)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Save and load the quantized model."""


# pylint:disable=import-error
import json
import os

import torch

try:
    import intel_extension_for_pytorch as ipex
except:
    assert False, "Please install IPEX for static quantization."

from neural_compressor.torch.utils import QCONFIG_NAME, WEIGHT_NAME, logger


def save(model, output_dir="./saved_results"):
    """Saves the quantized model to the output path.

    Args:
        model (RecursiveScriptModule): raw fp32 model or prepared model.
        output_dir (str, optional): output path to save the quantized model.
    """
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    qmodel_file_path = os.path.join(os.path.abspath(os.path.expanduser(output_dir)), WEIGHT_NAME)
    qconfig_file_path = os.path.join(os.path.abspath(os.path.expanduser(output_dir)), QCONFIG_NAME)
    device = next(model.parameters(), None).device.type if next(model.parameters(), None) else "cpu"
    if device == "cpu":
        model.ori_save(qmodel_file_path)
        with open(qconfig_file_path, "w") as f:
            json.dump(model.tune_cfg, f, indent=4)
    else:  # pragma: no cover
        from neural_compressor.common.utils import save_config_mapping

        torch.jit.save(model, qmodel_file_path)
        save_config_mapping(model.qconfig, qconfig_file_path)

    logger.info("Save quantized model to {}.".format(qmodel_file_path))
    logger.info("Save configuration of quantized model to {}.".format(qconfig_file_path))


def load(output_dir="./saved_results"):
    """Loads the quantized model from the output path.

    Args:
        output_dir (str, optional): output path to load the quantized model.

    Returns:
        A quantized model.
    """
    qmodel_file_path = os.path.join(os.path.abspath(os.path.expanduser(output_dir)), WEIGHT_NAME)
    model = torch.jit.load(qmodel_file_path)
    model = torch.jit.freeze(model.eval())
    logger.info("Quantized model loading successful.")
    return model
